#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2021/06/06 17:46
import re

from pyxllib.prog.pupil import check_install_package

check_install_package('joblib', 'joblib>=1.3.2')

from collections import OrderedDict
import sqlite3

from joblib import Parallel, delayed

from pyxllib.file.specialist.filelib import *
from pyxllib.file.specialist.dirlib import *
from pyxllib.file.specialist.download import *


def merge_jsonl(*infiles):
    data = []
    for f in infiles:
        data += XlPath(f).read_jsonl()
    return data


class JsonlDataFile:
    """ 通用的jsonl文件处理类 """

    def __init__(self, filepath=None, num_records=None):
        """
        从指定的jsonl文件中读取数据。可以选择读取全部数据或只读取前N条数据。

        :param str filepath: jsonl文件的路径
        :param int num_records: 指定读取的记录数量，如果为None则读取全部数据
        """
        self.infile = None
        self.records = []

        if filepath is not None:
            filepath = XlPath(filepath)
            if '?k' in filepath.name:  # 如果文件名中有'?'，则需要进行模式匹配检索
                new_name = filepath.name.replace('?k', '*')
                filepaths = list(filepath.parent.glob(new_name))
                if filepaths:
                    filepath = filepaths[0]  # 找到第1个匹配的文件
                    self.infile = XlPath(filepath)
            else:
                self.infile = filepath

        if self.infile and self.infile.is_file():  # 机制上文件也可能不存在的，有可能只是一个预设目录~
            if num_records is None:
                # 读取全部数据
                if self.infile.is_file():
                    self.records = self.infile.read_jsonl()
            else:
                # 只读取部分数据
                self.read_partial_records(num_records)

    def __len__(self):
        return len(self.records)

    def yield_record(self, start=0, end=None, step=1, batch_size=None):
        """ 返回指定区间的记录

        :param int start: 起始记录索引，默认为0
        :param int end: 结束记录索引，默认为None（读取到记录末尾）
        :param int step: 步长，默认为1
        :param int batch_size: 每批返回的记录数，如果为None，则逐记录返回
        """
        total_records = len(self.records)  # 获取总记录数

        # 处理负索引
        if start < 0 or (end is not None and end < 0):
            if start < 0:
                start = total_records + start
            if end is not None and end < 0:
                end = total_records + end

        iterator = islice(self.records, start, end, step)
        while True:
            batch = list(islice(iterator, batch_size))
            if not batch:
                break
            if batch_size is None:
                yield from batch
            else:
                yield batch

    def yield_group(self, key, sort_mode='keep'):
        """ 分组提取数据

        :param key: 一个函数，对record的映射，通过这个映射规则来分组
        :param sort_mode:
            keep: 保留原本的相对顺序
            id: 按照id的值进行排序
            sort: 按照key的值进行排序
        """
        # 1 创建一个默认字典来保存分组
        grouped_data = OrderedDict()

        records = self.records
        if sort_mode == 'id':
            records = sorted(records, key=lambda x: x['id'])

        # 2 对数据进行分组
        for record in records:
            k = key(record)
            if k not in grouped_data:
                grouped_data[k] = [record]
            else:
                grouped_data[k].append(record)

        # 3 将分组的数据重新排序并合并为一个新列表
        # 并且在这里可以进行一些分组信息的计算
        if sort_mode == 'sort':
            grouped_data = {k: grouped_data[k] for k in sorted(grouped_data.keys())}

        # 4 返回分组后的数据
        yield from grouped_data.values()

    def read_partial_records(self, num_records):
        """ 从jsonl文件中只读取指定数量的记录 """
        if self.infile and self.infile.is_file():
            try:
                lines = next(self.infile.yield_line(batch_size=num_records))
                for line in lines:
                    self.records.append(json.loads(line))
            except StopIteration:
                self.records = []

    def save(self, outfile=None, ensure_ascii=False, json_encoder=None):
        """ 将当前数据保存到指定的jsonl文件中 """
        if outfile is None:  # 默认保存回原文件
            outfile = self.infile
        p = XlPath(outfile)

        # 如果文件名包含'?k'，则替换'?'为self.records的数量
        if m := re.search(r'\?k', p.name):
            n = len(self.records)
            if n < 500:
                replace_str = f'{n}'  # 数量小于500，直接给出数量
            else:
                v = int(round(n / 1000))  # 数量大于等于500，以"千"为单位'k'，四舍五入计算
                replace_str = f'{v}k'
            # 用新字符串替换原来的字符串
            new_name = re.sub(r'\?k', replace_str, p.name)
            p = p.with_name(new_name)  # 更改文件名

        p.parent.mkdir(parents=True, exist_ok=True)
        p.write_jsonl(self.records, ensure_ascii=ensure_ascii, default=json_encoder)

    def browse_record(self, index=None, paths=None, **kwargs):
        """ 在浏览器中显示指定记录的内容 """
        from pyxllib.prog.specialist import browser

        # 如果未提供索引，则尝试使用查询参数找到第一个匹配的记录
        if index is None:
            index = self.find_index(paths, **kwargs)
            if index is None:
                raise ValueError('No matching record found')

        record = self.records[index]
        html_content = ['<html><body><pre>',
                        json.dumps(record, ensure_ascii=False, indent=4),
                        '</pre></body></html>']
        html_file = (XlPath.tempdir() / f'{self.__class__.__name__}_{index}.html')
        html_file.write_text('\n'.join(html_content))
        browser.html(html_file)

    def browse_records(self, indices=None, paths=None, **kwargs):
        """ 在浏览器中显示所有匹配的记录 """
        from pyxllib.prog.specialist import browser

        if indices is None:
            indices = list(self.find_indexs(paths, **kwargs))
        if not indices:
            raise ValueError('No matching records found')

        html_content = ['<html><body><h1>Matching Records: {}</h1>'.format(len(indices))]

        for index in indices:
            record = self.records[index]
            html_content.extend([
                '<h2>Record {}</h2>'.format(index),
                '<pre>',
                json.dumps(record, ensure_ascii=False, indent=4),
                '</pre>'
            ])

        html_content.append('</body></html>')
        html_file = (XlPath.tempdir() / f'{self.__class__.__name__}_matched.html')
        html_file.write_text('\n'.join(html_content))
        browser.html(html_file)

    def find_indexs(self, paths=None, **kwargs):
        """ 查找满足特定条件的记录的索引，返回所有匹配的结果 """
        paths = paths or {}

        for i, record in enumerate(self.records):
            # 检查kwargs中的所有条件
            for key, value in kwargs.items():
                if callable(value):
                    if not value(record.get(key)):
                        break
                elif record.get(key) != value:
                    break
            else:
                # 检查paths中的所有条件
                for path, value in paths.items():
                    try:
                        actual_value = eval(f'record{path}')
                    except Exception:
                        break

                    if callable(value):
                        if not value(actual_value):
                            break
                    elif actual_value != value:
                        break
                else:
                    # 如果记录满足所有条件，则返回它的索引
                    yield i

    def find_index(self, paths=None, **kwargs):
        """
        :param dict paths: 在比较复杂场景，无法使用kwargs定位的时候可以用这个规则
            key: 检索范式
            value: 需要满足的值
            示例： find_index({"['messages'][0]['role']": 'user'})
        :param kwargs: 直接子结点名称和对应的值
            示例：find_index(id=2023071320000003)

        补充说明：
            1、paths和kwargs可以组合使用，表示必须同时满足二者里限定的所有规则
            2、value可以写一个 def func(v)->bool的函数，输入对应的value，返回是否满足条件
        """
        return next(self.find_indexs(paths, **kwargs), None)

    def add_record_basic(self, **kwargs):
        """ 最基础的添加一个条目的接口 """
        record = kwargs
        self.records.append(record)
        return record

    @classmethod
    def read_from_files(cls, src_files):
        """ 从多个文件中读取并合并数据，并返回新的JsonlDataFile实例 """
        merged_records = []
        for file in src_files:
            jsonl_file = cls(file)
            merged_records.extend(jsonl_file.records)
        # 创建新的实例并返回
        new_instance = cls()
        new_instance.records = merged_records
        return new_instance

    @classmethod
    def read_from_dir(cls, src_dir):
        """ 从一个目录下的所有jsonl文件中读取并合并数据，并返回新的JsonlDataFile实例 """
        src_dir = XlPath(src_dir)
        src_files = [str(file_path) for file_path in src_dir.glob('*.jsonl')]
        return cls.read_from_files(src_files)

    def __add__(self, other):
        """ 实现类加法操作，合并两个JsonlDataFile的records """
        if not isinstance(other, JsonlDataFile):
            raise TypeError(f'Unsupported operand type: {type(other)}')
        result = JsonlDataFile()
        result.records = self.records + other.records
        return result

    def __iadd__(self, other):
        """ 实现原地加法操作，即 += """
        if not isinstance(other, JsonlDataFile):
            raise TypeError(f'Unsupported operand type: {type(other)}')
        self.records += other.records
        return self

    def process_each_record(self, func, *,
                            inplace=False,
                            timeout=None,
                            print_mode=0,
                            threads_num=1,
                            **kwargs):
        """ 对records中的每个record应用函数func，可以选择是否在原地修改，以及是否显示进度条

        :param function func: 对record进行处理的函数，应接受一个record作为参数并返回处理后的record，如果返回None则删除该record
        :param bool inplace: 是否在原地修改records，如果为False（默认），则创建新的JsonlDataFile并返回
        :param int print_mode: 是否显示处理过程的进度条，0表示不显示（默认），1表示显示
        :return JsonlDataFile or None: 如果inplace为False，则返回新的JsonlDataFile，否则返回None
        :param int threads_num: 线程数，默认为1，即单线程

        遍历self.records，对每个record执行func函数，如果func返回None，则不包含该record到新的records中。
        """
        backend = 'threading' if threads_num != 1 else 'sequential'

        if print_mode:
            parallel = Parallel(n_jobs=threads_num, backend=backend,
                                timeout=timeout, return_as='generator')
            tasks = [delayed(func)(record) for record in self.records]
            new_records = []
            for y in tqdm(parallel(tasks), total=len(self.records), **kwargs):
                if y:
                    new_records.append(y)
        else:
            parallel = Parallel(n_jobs=threads_num, backend=backend, timeout=timeout)
            tasks = [delayed(func)(record) for record in self.records]
            new_records = parallel(tasks)
            new_records = [y for y in new_records if y]

        if inplace:
            self.records = new_records

        return new_records

    def update_each_record(self, func,
                           timeout=None,
                           print_mode=0,
                           threads_num=1):
        """ 遍历并对原始数据进行更改 """
        return self.process_each_record(func,
                                        inplace=True,
                                        timeout=timeout,
                                        print_mode=print_mode,
                                        threads_num=threads_num)


class JsonlDataDir:
    """ 注意这个类开发目标，应该是尽量去模拟JsonDataFile，让下游工作更好衔接统一 """

    def __init__(self, root):
        """ 一般用来处理较大的jsonl文件，将其该放到一个目录里，拆分成多个jsonl文件

        注意待处理的文件名是依照 01.jsonl, 02.jsonl,... 的格式识别的，不要改动这个规则
        """
        self.root = XlPath(root)
        self.files = []
        self.update_subfiles()

    def update_subfiles(self):
        self.files = []
        for f in self.root.glob_files('*.jsonl'):
            if re.match(r'_?\d+$', f.stem):  # 目前先用'_?'兼容旧版，但以后应该固定只匹配_\d+
                self.files.append(f)

    def __bool__(self):
        if self.root.is_dir() and self.files:
            return True
        else:
            return False

    def count_records(self):
        total = 0
        for f in self.files:
            total += f.get_total_lines(skip_blank=True)
        return total

    def check(self, title=''):
        """ 检查一些数据状态 """
        print(title, '文件数:', len(self.files), '条目数:', self.count_records())

    @classmethod
    def init_from_file(cls, file, lines_per_file=10000):
        """ 从一个jsonl文件初始化一个JsonlDataDir对象 """
        file = XlPath(file)
        dst_dir = file.parent / file.stem
        if not dst_dir.is_dir() and file.is_file():
            file.split_to_dir(lines_per_file, dst_dir)
        c = cls(dst_dir)
        return c

    def _rearrange_group(self, lines_per_file=10000,
                         group_key=None, sort_mode='keep',
                         print_mode=1):
        # 1 使用sqlite3存储数据和分组信息
        # 创建一个临时文件来作为SQLite数据库
        temp_db_file = self.root / 'data.sqlite3'
        temp_db_file.delete()

        # 使用临时文件创建SQLite数据库连接
        conn = sqlite3.connect(temp_db_file)
        cursor = conn.cursor()

        # 创建一个临时表来存储jsonl数据
        cursor.execute('CREATE TABLE records (id INTEGER PRIMARY KEY AUTOINCREMENT,'
                       'data TEXT, group_key TEXT)')
        # 给group_key添加索引
        cursor.execute('CREATE INDEX idx_group_key ON records(group_key)')

        # 从jsonl文件加载数据到SQLite数据库
        commit_interval = 2000  # 多少记录执行一次commit
        count = 0
        for record in tqdm(self.yield_record(), desc='计算每个record分组', disable=not print_mode):
            count += 1
            group = group_key(record) if group_key else count
            group = str(group)
            cursor.execute('INSERT INTO records (data, group_key) VALUES (?, ?)',
                           (json.dumps(record, ensure_ascii=False), group))
            if count % commit_interval == 0:
                conn.commit()
        conn.commit()

        # 2 查询数据库以进行排序和分组，并将结果写入新的jsonl文件
        new_file_count = 0
        lines_written = 0
        current_file = None
        sort_sql = ''
        if sort_mode == 'id':
            sort_sql = 'ORDER BY id'
        elif sort_mode == 'sort':
            sort_sql = f'ORDER BY {group_key}'

        for group, in tqdm(cursor.execute('SELECT DISTINCT group_key FROM records').fetchall(),
                           desc='提取每一组数据',
                           disable=not print_mode):
            query = f'SELECT data FROM records WHERE group_key = ? {sort_sql}'
            cursor.execute(query, (group,))

            if current_file is None or lines_written >= lines_per_file:
                if current_file:
                    current_file.close()
                new_file_name = f'temp_{new_file_count}.jsonl'
                new_file_path = self.root / new_file_name
                current_file = new_file_path.open('w', encoding='utf-8')
                new_file_count += 1
                lines_written = 0

            while True:
                row = cursor.fetchone()
                if row is None:
                    break

                current_file.write(row[0] + '\n')
                lines_written += 1

        if current_file:
            current_file.close()

        # 3 关闭数据库连接并删除临时文件
        conn.close()
        temp_db_file.delete()

        # 4 删除旧文件，重命名新文件
        for f in self.files:
            f.delete()

        widths = len(str(new_file_count))
        for temp_file in self.root.glob('temp_*.jsonl'):
            n = int(re.search(r'\d+', temp_file.name).group())
            temp_file.rename(self.root / f'_{n:0{widths}}.jsonl')

    def rearrange(self, lines_per_file=10000, group_key=None,
                  sort_mode='keep', print_mode=1):
        """ 重新整理划分文件

        :param int lines_per_file: 每个文件的行数
        :param func group_key: 用来分组的函数，确保相同key的数据会被分到同一个文件里
        :param str sort_mode:
            keep: 保留原本的相对顺序
            id: 按照id的值进行排序
            sort: 按照key的值进行排序
        """
        if group_key is not None or sort_mode != 'keep':
            return self._rearrange_group(lines_per_file, group_key, sort_mode, print_mode)

        output_dir = self.root
        temp_prefix = 'temp_'

        new_file_count = 0
        new_file = None
        line_count = 0

        # 计算总行数以确定文件名的前导零数量
        total_lines = sum(1 for file in self.files for _ in file.open('r', encoding='utf-8'))
        num_digits = len(str((total_lines + lines_per_file - 1) // lines_per_file))

        for file in self.files:
            with file.open('r', encoding='utf-8') as f:
                for line in f:
                    if not line.strip():
                        continue
                    if line_count == 0:
                        if new_file is not None:
                            new_file.close()
                        new_file_name = f'{temp_prefix}{new_file_count:0{num_digits}d}.jsonl'
                        new_file_path = output_dir / new_file_name
                        new_file = new_file_path.open('w', encoding='utf-8')
                        new_file_count += 1

                    new_file.write(line)
                    line_count += 1

                    if line_count == lines_per_file:
                        line_count = 0

        if new_file is not None:
            new_file.close()

        # 删除旧文件
        for file in self.files:
            os.remove(file)

        # 将临时文件名更改为最终的文件名
        for temp_file in output_dir.glob(f'{temp_prefix}*.jsonl'):
            final_name = temp_file.name[len(temp_prefix) - 1:]
            temp_file.rename(output_dir / final_name)

    def yield_record(self, batch_size=None):
        """ 返回数据记录

        :param int batch_size: 每批返回的记录数，如果为None，则逐条返回
        """
        for i, file in enumerate(self.files):
            data = file.read_jsonl()
            iterator = iter(data)
            while True:
                batch = list(islice(iterator, batch_size))
                if not batch:
                    break
                if batch_size is None:
                    yield from batch
                else:
                    yield batch

    def yield_group(self, key, sort_mode='keep'):
        """ 分组提取数据

        :param key: 一个函数，对record的映射，通过这个映射规则来分组

        注意：这个分组只会对每个分文件单独执行，不会全局性质检索
            一般要用self.rearrange对全局的文件进行检索重排后再使用这个函数
        """
        for filepath in self.files:
            jdf = JsonlDataFile(filepath)
            yield from jdf.yield_group(key, sort_mode)

    def process_each_file(self, func=None, *,
                          print_mode=0, desc='process_each_file',
                          processes_num=1,
                          **kwargs):
        backend = 'loky' if processes_num != 1 else 'sequential'
        if print_mode:
            parallel = Parallel(n_jobs=processes_num, backend=backend, return_as='generator')
            tasks = [delayed(func)(file) for file in self.files]
            list(tqdm(parallel(tasks), total=len(self.files), desc=desc, **kwargs))
        else:
            parallel = Parallel(n_jobs=processes_num, backend=backend)
            tasks = [delayed(func)(file) for file in self.files]
            parallel(tasks)

    def process_each_record(self, func, *,
                            inplace=False, reset=False,
                            print_mode=2, desc=None,
                            timeout=None,
                            processes_num=1, threads_num=1,
                            dst_dir=None, json_encoder=None):
        """ 封装的对每个record进行操作的函数

        :param func: 外部传入的处理函数
        :param inplace: 是否修改原始数据
        :param reset: 是否重新处理已经处理过的文件
        :param print_mode:
            0 不显示
            1 只显示文件数进度
            2（默认） 显示更详细的文件内处理进度
        :param desc: print_mode=1的进度条标题
        :param timeout: 超时时间，但有些场合会使用不了（比如linux的子进程里不能使用singal）
            在用不了的场合，可以使用requests自带的timeout等各种机制来限时
        :param processes_num: 进程数，每个文件为单独一个进程
        :param threads_num: 线程数，每个文件处理的时候使用几个线程
        :param dst_dir: 要保存到的目标目录，未设置的时候不保存
        :param json_encoder: 有些不是标准的json数据结构，如何进行处理，有需要的时候一般会设置成str
        """
        files_num = len(self.files)

        def process_jsonl_file(srcfile):
            # 1 如果没有reset，且dstfile存在，则不处理
            srcfile = XlPath(srcfile)
            if dst_dir:
                dstfile = XlPath(dst_dir) / srcfile.name
            else:
                dstfile = None
            if not reset and dstfile and dstfile.is_file():
                return

            # 2 跑特定文件里的条目
            jdf = JsonlDataFile(srcfile)
            new_records = jdf.process_each_record(func,
                                                  inplace=inplace,
                                                  print_mode=print_mode == 2,
                                                  desc=f'{jdf.infile.name}/{files_num}',
                                                  timeout=timeout,
                                                  threads_num=threads_num,
                                                  mininterval=processes_num * 3,
                                                  )

            # 3 是否修改原文件，是否保存到dst_dir
            if inplace:
                jdf.save()

            if dstfile:
                jdf = JsonlDataFile()
                jdf.records = new_records
                jdf.save(dstfile, json_encoder=json_encoder)

        self.process_each_file(process_jsonl_file,
                               processes_num=processes_num,
                               print_mode=print_mode == 1, desc=desc)

    def process_each_group(self, func, group_key, sort_mode='keep', *,
                           inplace=False, reset=False,
                           print_mode=1, desc=None,
                           processes_num=1,
                           dst_dir=None,
                           json_encoder=None):
        """ 封装的对每组records的处理

        todo 230909周六14:00，还有些细节功能可能不完善，比如内部的进度条，多线程等，等后续使用的时候慢慢优化
        """

        def process_jsonl_file(srcfile):
            # 1 如果没有reset，且dstfile存在，则不处理
            srcfile = XlPath(srcfile)
            if dst_dir:
                dstfile = XlPath(dst_dir) / srcfile.name
            else:
                dstfile = None
            if not reset and dstfile and dstfile.is_file():
                return

            # 2 跑特定文件里的条目
            jdf = JsonlDataFile(srcfile)
            new_records = []
            for records in jdf.yield_group(group_key, sort_mode):
                records2 = func(records)
                if records2:
                    new_records.extend(records2)

            # 3 是否修改原文件，是否保存到dst_dir
            if inplace:
                jdf.records = new_records
                jdf.save()

            if dstfile:
                jdf = JsonlDataFile()
                jdf.records = new_records
                jdf.save(dstfile, json_encoder=json_encoder)

        self.process_each_file(process_jsonl_file,
                               processes_num=processes_num,
                               print_mode=print_mode == 1, desc=desc)

    def save(self, dst_path=None):
        """ 将数据合并到一个jsonl文件中 """
        if not dst_path:
            dst_path = self.root.parent / f'{self.root.name}.jsonl'
        dst_path = XlPath(dst_path)
        dst_path.parent.mkdir(parents=True, exist_ok=True)
        with dst_path.open('w', encoding='utf8') as f:
            for file in tqdm(self.files, desc=f'合并文件并保存 {dst_path.name}'):
                with file.open('r', encoding='utf8') as f2:
                    for line in f2:
                        if line.strip():  # 不存储空行
                            f.write(line)

    def clear(self):
        for f in self.files:
            f.delete()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2023/09/16

from pyxllib.prog.pupil import check_install_package
# check_install_package('sklearn', 'scikit-learn')

# 这个需要C++14编译器 https://download.microsoft.com/download/5/f/7/5f7acaeb-8363-451f-9425-68a90f98b238/visualcppbuildtools_full.exe
# 在需要的时候安装，防止只是想用pyxllib很简单的功能，但是在pip install阶段处理过于麻烦
# 字符串计算编辑距离需要
# check_install_package('Levenshtein', 'python-Levenshtein')

from collections import defaultdict
import warnings

warnings.filterwarnings("ignore", message="loaded more than 1 DLL from .libs:")
warnings.filterwarnings("ignore", category=FutureWarning,
                        module="sklearn.cluster._agglomerative",
                        lineno=1005)

from more_itertools import chunked

from tqdm import tqdm

try:
    import numpy as np
except ModuleNotFoundError:
    pass

try:  # 如果不使用字符串编辑距离相关的功能，那这个包导入失败也没关系
    import Levenshtein
except ModuleNotFoundError:
    pass

try:  # 层次聚类相关的功能
    from sklearn.cluster import AgglomerativeClustering
except ModuleNotFoundError:
    pass


class DataMatcher:
    """ 泛化的匹配类，对任何类型的数据进行匹配 """

    def __init__(self, *, cmp_key=None):
        """
        :param cmp_key: 当设置该值时，表示data中不是整个用于比较，而是有个索引列
        """
        self.cmp_key = cmp_key
        self.data = []  # 用于匹配的数据

    def __getitem__(self, i):
        return self.data[i]

    def __delitem__(self, i):
        del self.data[i]

    def __len__(self):
        return len(self.data)

    def compute_similarity(self, x, y):
        """ 计算两个数据之间的相似度，这里默认对字符串使用编辑距离 """
        if self.cmp_key:
            x = x[self.cmp_key]
        ratio = Levenshtein.ratio(x, y)
        return ratio

    def add_candidate(self, data):
        """添加候选数据"""
        self.data.append(data)

    def find_best_matches(self, item, top_n=1, print_mode=0):
        """ 找到与给定数据项最匹配的候选项。

        :param item: 需要匹配的数据项。
        :param top_n: 返回的最佳匹配数量。
        :return: 一个包含(index, similarity)的元组列表，代表最佳匹配。
        """
        # 计算所有候选数据的相似度
        similarities = [(i, self.compute_similarity(candidate, item))
                        for i, candidate in tqdm(enumerate(self.data), disable=not print_mode)]

        # 按相似度降序排序
        sorted_matches = sorted(similarities, key=lambda x: x[1], reverse=True)

        return sorted_matches[:top_n]

    def find_best_match_items(self, item, top_n=1):
        """ 直接返回匹配的数据内容，而不是下标和相似度 """
        matches = self.find_best_matches(item, top_n=top_n)
        return [self.data[m[0]] for m in matches]

    def find_best_match(self, item):
        """ 返回最佳匹配 """
        matches = self.find_best_matches(item, top_n=1)
        return matches[0]

    def find_best_match_item(self, item):
        """ 直接返回匹配的数据内容，而不是下标和相似度 """
        items = self.find_best_match_items(item)
        return items[0]

    def agglomerative_clustering(self, threshold=0.5):
        """ 对内部字符串进行层次聚类

        :param threshold: 可以理解成距离的阈值，距离小于这个阈值的字符串会被聚为一类
            值越小，分出的类别越多越细
        """
        # 1 给每个样本标类别
        distance_matrix = np.zeros((len(self), len(self)))
        for i in range(len(self)):
            for j in range(i + 1, len(self)):
                # 我们需要距离，所以用1减去相似度
                distance = 1 - self.compute_similarity(self.data[i], self.data[j])
                distance_matrix[i, j] = distance_matrix[j, i] = distance

        # 进行层次聚类
        clustering = AgglomerativeClustering(n_clusters=None, affinity='precomputed',
                                             distance_threshold=threshold,
                                             linkage='complete')
        labels = clustering.fit_predict(distance_matrix)

        return labels

    def display_clusters(self, threshold=0.5):
        """ 根据agglomerative_clustering的结果，显示各个聚类的内容 """
        labels = self.agglomerative_clustering(threshold=threshold)
        cluster_dict = defaultdict(list)

        # 组织数据到字典中
        for idx, label in enumerate(labels):
            cluster_dict[label].append(self.data[idx])

        # 按标签排序并显示
        result = {}
        for label, items in sorted(cluster_dict.items(), key=lambda x: -len(x[1])):
            result[label] = items

        return result

    def get_center_sample(self, indices=None):
        """ 获取一个数据集的中心样本

        :param indices: 数据项的索引列表。如果为None，则考虑所有数据。
        :return: 中心样本的索引。
        """
        if indices is None:
            indices = range(len(self.data))

        cached_results = {}

        def get_similarity(i, j):
            """ 获取两个索引的相似度，利用缓存来避免重复计算 """
            if (i, j) in cached_results:
                return cached_results[(i, j)]
            sim_val = self.compute_similarity(self.data[indices[i]], self.data[indices[j]])
            cached_results[(i, j)] = cached_results[(j, i)] = sim_val
            return sim_val

        center_idx = max(indices, key=lambda x: sum(get_similarity(x, y) for y in indices))
        return center_idx


class GroupedDataMatcher(DataMatcher):
    """ 对数据量特别大的情况，我们可以先对数据进行分组，然后再对每个分组进行匹配 """

    def __init__(self):
        """ 初始化一个分组数据匹配器 """
        super().__init__()
        self.groups = dict()

    def _sort_groups(self):
        """ 按照组员数量从多到少排序groups """
        new_groups = {}
        for rep, items in sorted(self.groups.items(), key=lambda x: -len(x[1])):
            new_groups[rep] = items
        self.groups = new_groups

    def merge_group(self, indices, threshold=0.5, strategy='center'):
        """ 对输入的索引进行合并，根据阈值生成分组

        :param indices: 数据项的索引列表。
        :param threshold: 两个数据项的距离小于此阈值时，它们被认为是相似的。
        :param strategy: 选择组代表的策略，可以是'center'或'first'。
        :return: 一个字典，键是代表性数据项的索引，值是相似数据项的索引列表。
        """
        # 1 给每个样本标类别
        n = len(indices)
        if n == 1:
            return {indices[0]: indices}

        distance_matrix = np.zeros((n, n))
        for i in range(n):
            for j in range(i + 1, n):
                distance = 1 - self.compute_similarity(self.data[indices[i]], self.data[indices[j]])
                distance_matrix[i, j] = distance_matrix[j, i] = distance

        clustering = AgglomerativeClustering(n_clusters=None, affinity='precomputed',
                                             distance_threshold=threshold,
                                             linkage='average')
        labels = clustering.fit_predict(distance_matrix)

        # 2 分组字典
        cluster_dict = defaultdict(list)
        for i, label in enumerate(labels):
            cluster_dict[label].append(indices[i])

        # 3 改成代表样本映射到一组里，并且按照样本数从多到少排序
        result = {}
        for label, items in sorted(cluster_dict.items(), key=lambda x: -len(x[1])):
            if strategy == 'first':
                representative = items[0]
            elif strategy == 'center':
                local_indices = [i for i, idx in enumerate(indices) if idx in items]
                sub_matrix = distance_matrix[np.ix_(local_indices, local_indices)]
                avg_distances = sub_matrix.mean(axis=1)
                representative_idx = np.argmin(avg_distances)
                representative = items[representative_idx]
            else:
                raise ValueError(f'Invalid strategy: {strategy}')
            result[representative] = items

        return result

    def init_groups(self, threshold=0.5, batch_size=1000, print_mode=0):
        """ 初始化数据的分组

        :param threshold: 两个数据项的距离小于此阈值时，它们被认为是相似的。
            这里写成1的话，一般就是故意特地把类别只分成一类
        :param batch_size: 由于数据可能很大，可以使用批量处理来减少计算量。
        :return: 一个字典，键是代表性数据项的索引，值是相似数据项的索引列表。
        """
        # 1 最开始每个样本都是一个组
        groups = {i: [i] for i in range(len(self.data))}
        new_groups = {}

        # 2 不断合并，直到没有组数变化
        while len(groups) > 1:
            for indices in chunked(groups.keys(), batch_size):
                # 对于这里返回的字典，原groups里的values也要对应拼接的
                indices2 = self.merge_group(indices, threshold=threshold)
                for idx, idxs in indices2.items():
                    # 获取原始分组中的索引
                    original_idxs = [groups[original_idx] for original_idx in idxs]
                    # 展平列表并分配到新分组中
                    new_groups[idx] = [item for sublist in original_idxs for item in sublist]

            # 如果分组没有发生变化，退出循环
            if len(new_groups) == len(groups):
                break

            if print_mode:
                print(f'Groups number: {len(new_groups)}')

            groups = new_groups
            new_groups = {}

        self.groups = groups
        self._sort_groups()
        return self.groups

    def split_large_groups(self, max_group_size, threshold=0.5):
        """ 对于样本数过多的类，进行进一步的拆分

        :param max_group_size: 一个组内的最大样本数，超过这个数就会被进一步拆分。
        :param threshold: 用于拆分的阈值，两个数据项的距离小于此阈值时，它们被认为是相似的。
        :return: 返回拆分后的分组。
        """

        refined_groups = {}
        for rep, items in self.groups.items():
            if len(items) > max_group_size:
                # 该组样本数超过阈值，需要进一步拆分
                sub_groups = self.merge_group(items, threshold)
                refined_groups.update(sub_groups)
            else:
                # 该组样本数在阈值范围内，保持不变
                refined_groups[rep] = items

        self.groups = refined_groups
        self._sort_groups()
        return refined_groups

    def merge_small_groups(self, min_group_size=10):
        """ 将样本数较小的组合并成一个大组

        :param min_group_size: 一个组的最小样本数，低于这个数的组将被合并。
        :return: 返回合并后的分组。
        """

        merged_group = []
        preserved_groups = {}

        for rep, items in self.groups.items():
            if len(items) < min_group_size:
                # 该组样本数低于阈值，将其添加到待合并的大组中
                merged_group.extend(items)
            else:
                # 该组样本数大于等于阈值，保留原状
                preserved_groups[rep] = items

        if merged_group:
            rep_item = self.merge_group(merged_group, 1)
            for rep, items in rep_item.items():
                preserved_groups[rep] = items

        self.groups = preserved_groups
        self._sort_groups()
        return preserved_groups



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2022/06/09 16:26

"""
针对PostgreSQL封装的工具
"""

import os
import sys

from pyxllib.prog.pupil import check_install_package
from pyxllib.file.specialist import XlPath, ensure_localdir

# 使用binary安装，好像就不需要自己配置依赖了
# if sys.platform == 'win32':
#     # windows系统中pg的默认安装位置
#     # https://www.yuque.com/xlpr/pyxllib/install_psycopg
#     pgdll_dir = XlPath('C:/Program Files/PostgreSQL/14/bin')
#     if not pgdll_dir.is_dir():  # 换一个自定义目录继续检查
#         pgdll_dir = XlPath.userdir() / '.xlpr/pgdll'
#     if not pgdll_dir.is_dir():
#         print('没有PostgreSQL对应的dll文件，自动下载：')
#         ensure_localdir(pgdll_dir, 'https://xmutpriu.com/download/pgdll_v14.zip')
#     os.environ['PATH'] += ';' + pgdll_dir.as_posix()

check_install_package('psycopg', 'psycopg[binary]')

import io
from collections import Counter
import json
import json
import textwrap
import datetime

from tqdm import tqdm

import psycopg
import psycopg.rows

from pyxllib.prog.newbie import round_int
from pyxllib.prog.pupil import utc_now, utc_timestamp
from pyxllib.prog.specialist import XlOsEnv
from pyxllib.file.specialist import get_etag
from pyxllib.data.sqlite import SqlBase


class Connection(psycopg.Connection, SqlBase):

    def __1_库(self):
        pass

    def __2_表格(self):
        pass

    def get_table_names(self):
        # cmd = "SELECT table_name FROM information_schema.tables WHERE table_schema='public' AND table_type='BASE TABLE'"
        cmd = "SELECT tablename FROM pg_tables WHERE schemaname='public'"
        return [x[0] for x in self.execute(cmd)]

    def has_table(self, table_name, schemaname='public'):
        query = ["SELECT EXISTS (SELECT FROM pg_tables ",
                 f"WHERE schemaname='{schemaname}' AND tablename='{table_name}')"]
        res = self.execute(' '.join(query))
        return res.fetchone()[0]

    def get_column_names(self, table_name):
        """ 【查】表格有哪些字段
        """
        cmd = f"SELECT column_name FROM information_schema.columns WHERE table_name='{table_name}'"
        return list(self.exec_col(cmd))

    def ensure_column(self, table_name, col_name, *args, comment=None, **kwargs):
        super(Connection, self).ensure_column(table_name, col_name, *args, **kwargs)
        if comment:
            self.set_column_comment(table_name, col_name, comment)

    def set_column_comment(self, table_name, col_name, comment):
        # 这里comment按%s填充会报错。好像是psycopg库的问题，这种情况因为没有表格字段参照类型，会不知道要把comment转成字符串。
        # 然后使用py的f字符串填充的话，要避免comment有引号等特殊字符，就用了特殊的转义标记$zyzf$
        # 暂时找不到更优雅的方式~~ 而且这个问题，可能以后其他特殊的sql语句也会遇到。。。
        self.execute(f"COMMENT ON COLUMN {table_name}.{col_name} IS $zyzf${comment}$zyzf$")
        self.commit()

    def __3_execute(self):
        pass

    def exec_nametuple(self, *args, **kwargs):
        cur = self.cursor(row_factory=psycopg.rows.namedtuple_row)
        data = cur.execute(*args, **kwargs)
        # cur.close()
        return data

    def exec_dict(self, *args, **kwargs):
        cur = self.cursor(row_factory=psycopg.rows.dict_row)
        data = cur.execute(*args, **kwargs)
        # cur.close()
        return data

    def __4_数据类型(self):
        pass

    @classmethod
    def cvt_type(cls, val):
        if isinstance(val, dict):
            val = json.dumps(val, ensure_ascii=False)
        # 注意list数组类型读、写都会自动适配py
        return val

    @classmethod
    def autotype(cls, val):
        if isinstance(val, str):
            return 'text'
        elif isinstance(val, int):
            return 'int4'  # int2、int8
        elif isinstance(val, bool):
            return 'boolean'
        elif isinstance(val, float):
            return 'float4'
        elif isinstance(val, dict):
            return 'jsonb'
        else:  # 其他list等类型，可以用json.dumps或str转文本存储
            return 'text'

    def __5_增删改查(self):
        pass

    def insert_row(self, table_name, cols, *, on_conflict='DO NOTHING', commit=True):
        """ 【增】插入新数据

        :param dict cols: 用字典表示的要插入的值
        :param on_conflict: 如果已存在的处理策略
            DO NOTHING，跳过不处理
            REPLACE，这是个特殊标记，会转为主键冲突也仍然更新所有值
            (id) DO UPDATE SET host_name=EXCLUDED.host_name, nick_name='abc'
                也可以写复杂的处理算法规则，详见 http://postgres.cn/docs/12/sql-insert.html
                比如这里是插入的id重复的话，就把host_name替换掉，还可以指定nick_name替换为'abc'
                注意前面的(id)是必须要输入的
        """
        ks = ','.join(cols.keys())
        vs = ','.join(['%s'] * (len(cols.keys())))
        query = f'INSERT INTO {table_name}({ks}) VALUES ({vs})'
        params = self.cvt_types(cols.values())

        if on_conflict == 'REPLACE':
            on_conflict = f'ON CONFLICT ON CONSTRAINT {table_name}_pkey DO UPDATE SET ' + \
                          ','.join([f'{k}=EXCLUDED.{k}' for k in cols.keys()])
        else:
            on_conflict = f'ON CONFLICT {on_conflict}'
        query += f' {on_conflict}'

        self.execute(query, params)
        if commit:
            self.commit()

    def insert_rows(self, table_name, keys, ls, *, on_conflict='DO NOTHING', commit=True):
        """ 【增】插入新数据

        :param str keys: 要插入的字段名，一个字符串，逗号,隔开属性值
        :param list[list] ls: n行m列的数组

        >> con.insert_rows('hosts2', 'id,host_name,nick_name', [[1, 'test5', 'dcba'], [11, 'test', 'aabb']])
        """
        n, m = len(ls), len(keys.split(','))
        vs = ','.join(['%s'] * m)
        query = f'INSERT INTO {table_name}({keys}) VALUES ' + ','.join([f'({vs})'] * n)
        params = []
        for cols in ls:
            params += self.cvt_types(cols)

        if on_conflict == 'REPLACE':
            on_conflict = f'ON CONFLICT ON CONSTRAINT {table_name}_pkey DO UPDATE SET ' + \
                          ','.join([f'{k.strip()}=EXCLUDED.{k.strip()}' for k in keys.split(',')])
        else:
            on_conflict = f'ON CONFLICT {on_conflict}'
        query += f' {on_conflict}'

        self.execute(query, params)
        if commit:
            self.commit()


"""
【关于为什么XlprDb要和pglib合一个文件】
好处
1、代码集中管理，方便开发效率

坏处 - 对治
1、导入代码多，效率低
    - 导入代码其实很快的，不差这一点点
2、仅用简单功能的时候，会引入多余复杂功能
    - 如果单拉一个xlprdb.py文件，依然不能头部import复杂库，因为XlprDb也存在需要精简使用的场景
        既然都要有精简的部分，干脆都把代码弄到一起得了
        然后复杂包都在成员函数里单独import
"""


class XlprDb(Connection):
    """ xlpr统一集中管理的一个数据库

    为了一些基础的数据库功能操作干净，尽量不要在全局导入过于复杂的包，可以在每个特定功能里导特定包
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.seckey = ''

    @classmethod
    def set_conninfo(cls, conninfo, seckey=''):
        """ 提前将登录信息加密写到环境变量中，这样代码中就没有明文的账号密码信息

        :param conninfo:
        :param seckey: 如果要获得数据库里较重要的密码等信息，需要配置key值，否则默认可以不设

        使用后，需要重启IDE重新加载环境变量
        并且本句明文代码需要删除
        """
        # TODO 目前只设一个账号，后续可以扩展支持多个账号指定配置
        # conninfo = 'postgresql://postgres:yourpassword@172.16.170.110/xlpr'
        return XlOsEnv.persist_set('XlprDbAccount', {'conninfo': conninfo, 'seckey': seckey}, encoding=True)

    @classmethod
    def connect(cls, conninfo='', seckey='', *,
                autocommit=False, row_factory=None, context=None, **kwargs) -> 'XlprDb':
        """ 因为要标记 -> XlprDb，IDE才会识别出类别，有自动补全功能
        但在类中写@classmethod，无法标记 -> XlprDb，所以就放外面单独写一个方法了
        """
        d = XlOsEnv.get('XlprDbAccount', decoding=True)
        if conninfo == '':
            conninfo = d['conninfo']
        if seckey == '' and isinstance(d, dict) and 'seckey' in d:
            seckey = d['seckey']
        # 注意这里获取的是XlprDb类型
        con = super(XlprDb, cls).connect(conninfo, autocommit=autocommit, row_factory=row_factory, context=context,
                                         **kwargs)
        con.seckey = seckey
        return con

    @classmethod
    def connect2(cls, name, passwd, database=None, ip_list=None, **kwargs):
        """ 简化的登录方式，并且自带优先尝试局域网连接，再公网连接

        一般用在jupyter等对启动速度没有太高要求的场合，因为不断尝试localhost等需要耗费不少时间
        """
        if database is None:
            database = name

        xldb = None
        if ip_list is None:
            ip_list = ['localhost', '172.16.170.136', 'xmutpriu.com']
        for ip in ip_list:
            try:
                xldb = XlprDb.connect(f'postgresql://{name}:{passwd}@{ip}/{database}', **kwargs)
                break
            except psycopg.OperationalError:
                pass

        return xldb

    def __1_hosts相关数据表操作(self):
        pass

    def update_host(self, host_name, accounts=None, **kwargs):
        """ 更新一台服务器的信息

        :param dict accounts: 账号信息，记得要列全，比如
            {'root': '123456', 'chenkunze': '654321'}
        """
        if not self.execute('SELECT EXISTS (SELECT FROM hosts WHERE host_name=%s)', (host_name,)).fetchone()[0]:
            self.insert_row('hosts', {'host_name': host_name})
        if kwargs:
            self.update_row('hosts', kwargs, {'host_name': host_name})
        if accounts:
            self.execute('UPDATE hosts SET accounts=pgp_sym_encrypt(%s, %s) WHERE host_name=%s',
                         (json.dumps(accounts, ensure_ascii=False), self.seckey, host_name))
        self.commit()

    def set_host_account(self, host_name, user_name, passwd):
        """ 设置某台服务器的一个账号密码

        >> xldb.set_host_account('titan2', 'chenkunze', '123456')

        """
        # 读取旧的账号密码字典数据
        d = self.execute("SELECT pgp_sym_decrypt(accounts, %s)::jsonb FROM hosts WHERE host_name=%s",
                         (self.seckey, host_name)).fetchone()[0]
        # 修改字典数据
        d[user_name] = str(passwd)
        # 将新的字典数据写回数据库
        self.execute('UPDATE hosts SET accounts=pgp_sym_encrypt(%s, %s) WHERE host_name=%s',
                     (json.dumps(d, ensure_ascii=False), self.seckey, host_name))
        self.commit()

    def login_ssh(self, host_name, user_name, map_path=None, **kwargs) -> 'XlSSHClient':
        """ 通过数据库里的服务器数据记录，直接登录服务器 """
        from pyxllib.ext.unixlib import XlSSHClient

        if host_name.startswith('g_'):
            host_ip = self.execute("SELECT host_ip FROM hosts WHERE host_name='xlpr0'").fetchone()[0]
            pw, port = self.execute('SELECT (pgp_sym_decrypt(accounts, %s)::jsonb)[%s]::text, frpc_port'
                                    ' FROM hosts WHERE host_name=%s',
                                    (self.seckey, user_name, host_name[2:])).fetchone()
        else:
            port = 22
            host_ip, pw = self.execute('SELECT host_ip, (pgp_sym_decrypt(accounts, %s)::jsonb)[%s]::text'
                                       ' FROM hosts WHERE host_name=%s',
                                       (self.seckey, user_name, host_name)).fetchone()

        if map_path is None:
            if sys.platform == 'win32':
                map_path = {'C:/': '/'}
            else:
                map_path = {'/': '/'}

        return XlSSHClient(host_ip, user_name, pw[1:-1], port=port, map_path=map_path, **kwargs)

    def __2_xlapi相关数据表操作(self):
        """
        files，存储二进制文件的表
            etag，文件、二进制对应的etag值
            meta，可以存储不同数据类型一些特殊的属性，比如图片可以存储高、宽
                但因为用户使用中，为了提高速度，以及减少PIL等依赖，执行中不做计算
                可以其他途径使用定期脚本自动处理
        xlapi，底层api调用的记录统计
            input，所有输入的参数
                mode，使用的算法接口
                etag，涉及到大文件数据的，以打包后的文件的etag做记录
            output，运行结果
            elapse_ms，调用函数的用时，不含后处理转换格式的时间
        xlserver，服务端记录的使用情况
        """
        pass

    def insert_row2files(self, buffer, *, etag=None, **kwargs):
        """

        为了运算效率考虑，除了etag需要用于去重，是必填字段
        其他字段默认先不填充计算
        """
        # 1 已经有的不重复存储
        if etag is None:
            etag = get_etag(buffer)

        res = self.execute('SELECT etag FROM files WHERE etag=%s', (etag,)).fetchone()
        if res:  # 已经做过记录的，不再重复记录
            return

        # 2 没有的图，做个备份
        kwargs['etag'] = etag
        kwargs['data'] = buffer
        kwargs['fsize_kb'] = round_int(len(buffer) / 1024)

        self.insert_row('files', kwargs)
        self.commit()

    def get_xlapi_record(self, **input):
        res = self.execute('SELECT id, output FROM xlapi WHERE input=%s', (self.cvt_type(input),)).fetchone()
        if res:
            _id, output = res
            output['xlapi_id'] = _id
            return output

    def insert_row2xlapi(self, input, output, elapse_ms, *, on_conflict='(input) DO NOTHING'):
        """ 往数据库记录当前操作内容

        :return: 这个函数比较特殊，需要返回插入的条目的id值
        """
        if on_conflict == 'REPLACE':
            on_conflict = "(input) DO UPDATE " \
                          "SET output=EXCLUDED.output, elapse_ms=EXCLUDED.elapse_ms, update_time=EXCLUDED.update_time"

        input = json.dumps(input, ensure_ascii=False)
        self.insert_row('xlapi', {'input': input, 'output': output,
                                  'elapse_ms': elapse_ms, 'update_time': utc_timestamp(8)},
                        on_conflict=on_conflict)

        return self.execute('SELECT id FROM xlapi WHERE input=%s', (input,)).fetchone()[0]

    def insert_row2xlserver(self, request, xlapi_id=0, **kwargs):
        kw = {'remote_addr': request.headers.get('X-Real-IP', request.remote_addr),
              'route': '/'.join(request.base_url.split('/')[3:]),
              'update_time': utc_timestamp(8),
              'xlapi_id': xlapi_id}
        kw.update(kwargs)
        print(kw)  # 监控谁在用api
        self.insert_row('xlserver', kw)

    def __3_host_trace相关可视化(self):
        """ TODO dbview 改名 host_trace """
        pass

    def __dbtool(self):
        pass

    def record_host_usage(self, cpu=True, gpu=True, disk=False):
        """ 记录服务器各种状况，存储到PG数据库

        TODO 并行处理
        TODO 功能还可以增加：gpu显卡温度、硬盘读写速率检查、网络上传下载带宽
        """
        # 1 服务器列表
        host_names = list(self.exec_col('SELECT host_name FROM hosts WHERE id > 1 ORDER BY id'))
        host_cpu_gb = {h: v for h, v in self.execute('SELECT host_name, cpu_gb FROM hosts')}

        # 2 去所有服务器取使用情况
        for i, host_name in enumerate(host_names, start=1):
            print('-' * 20, i, host_name, '-' * 20)
            try:
                ssh = self.login_ssh(host_name, 'root', relogin=5, relogin_interval=0.2)
                status = {}
                if cpu:
                    data = ssh.check_cpu_usage(print_mode=True)
                    status['cpu'] = {k: v[0] for k, v in data.items()}
                    status['cpu_memory'] = {k: round(v[1] * host_cpu_gb[host_name] / 100, 2) for k, v in data.items()}
                if gpu:
                    status['gpu_memory'] = ssh.check_gpu_usage(print_mode=True)
                if disk:
                    # 检查磁盘空间会很慢，如果超时可以跳过。
                    status['disk_memory'] = ssh.check_disk_usage(print_mode=True, timeout=7200)
            except Exception as e:
                status = {'error': f'{str(type(e))[8:-2]}: {e}'}
                print(status)

            if status:
                self.insert_row('host_trace',
                                {'host_name': host_name, 'status': status, 'update_time': utc_timestamp(8)})
            print()

    def _get_host_trace_total(self, mode, title, yaxis_name, date_trunc, recent, host_attr):
        # CREATE INDEX ON gpu_trace (update_time);  -- update_time最好建个索引
        ls = self.execute(textwrap.dedent(f"""\
        WITH cte1 AS (  -- 筛选近期内的数据，并且时间做trunc处理
            SELECT host_name, (status)['{mode}'], date_trunc('{date_trunc}', update_time) ttime
            FROM host_trace WHERE update_time > %s AND (status ? '{mode}')
        ), cte2 AS (  -- 每个时间里每个服务器的多条记录取平均
            SELECT ttime, cte1.host_name, jsonb_div(jsonb_deep_sum(status), count(*)) status
            FROM cte1 GROUP BY ttime, cte1.host_name
        )  -- 接上每台服务器显存总值，并且分组得到每个时刻总情况
        SELECT ttime, {host_attr}, jsonb_deep_sum(status)
        FROM cte2 JOIN hosts ON cte2.host_name = hosts.host_name
        GROUP BY ttime ORDER BY ttime"""), ((utc_now(8) - recent).isoformat(timespec='seconds'),)).fetchall()
        return self._create_stack_chart(title, ls, yaxis_name=yaxis_name)

    def _get_host_trace_per_host(self, hostname, mode, title, yaxis_name, date_trunc, recent, host_attr):
        ls = self.execute(textwrap.dedent(f"""\
        WITH cte1 AS (
            SELECT (status)['{mode}'], date_trunc('{date_trunc}', update_time) ttime
            FROM host_trace WHERE host_name='{hostname}' AND update_time > %s AND (status ? '{mode}')
        ), cte2 AS (
            SELECT ttime, jsonb_div(jsonb_deep_sum(status), count(*)) status
            FROM cte1 GROUP BY ttime
        )
        SELECT ttime, {host_attr}, jsonb_deep_sum(status)
        FROM cte2 JOIN hosts ON hosts.host_name='{hostname}'
        GROUP BY ttime ORDER BY ttime"""), ((utc_now(8) - recent).isoformat(timespec='seconds'),)).fetchall()
        if not ls:  # 有的服务器可能数据是空的
            return '', 0
        return self._create_stack_chart(title, ls, yaxis_name=yaxis_name)

    def _create_stack_chart(self, title, ls, *, yaxis_name=''):
        """ 创建展示表

        :param title: 表格标题
        :param list ls: n*3，第1列是时间，第2列是总值，第3列是每个用户具体的数据
        """
        from pyecharts.charts import Line
        from pyxllib.data.echarts import get_render_body

        map_user_name = {}
        for ks, v in self.execute('SELECT account_names, name FROM users'):
            for k in ks:
                map_user_name[k] = v

        # 1 计算涉及的所有用户以及使用总量
        all_users_usaged = Counter()
        last_time = None
        for x in ls:
            hours = 0 if last_time is None else ((x[0] - last_time).total_seconds() / 3600)
            last_time = x[0]
            for k, v in x[2].items():
                if k == '_total':
                    continue
                all_users_usaged[map_user_name.get(k, k)] += v * hours

        # ls里的姓名也要跟着更新
        for i, x in enumerate(ls):
            ct = Counter()
            for k, v in x[2].items():
                ct[map_user_name.get(k, k)] += v
            ls[i] = (x[0], x[1], ct)

        # 2 转图表可视化
        def to_list(values):
            return [(x[0], v) for x, v in zip(ls, values)]

        def pretty_val(v):
            return round_int(v) if v > 100 else round(v, 2)

        chart = Line()
        chart.set_title(title)
        chart.options['xAxis'][0].update({'min': ls[0][0], 'type': 'time',
                                          # 'minInterval': 3600 * 1000 * 24,
                                          'name': '时间', 'nameGap': 50, 'nameLocation': 'middle'})
        chart.options['yAxis'][0].update({'name': yaxis_name, 'nameGap': 50, 'nameLocation': 'middle'})
        # 目前是比较暴力的方法调整排版，后续要研究是不是能更自动灵活些
        chart.options['legend'][0].update({'top': '6%', 'icon': 'pin'})
        chart.options['grid'] = [{'top': 55 + len(all_users_usaged) * 4, 'containLabel': True}]
        chart.options['tooltip'].opts.update({'axisPointer': {'type': 'cross'}, 'trigger': 'item'})

        chart.add_series(f'total{pretty_val(ls[0][1]):g}', to_list([x[1] for x in ls]), areaStyle={})
        for user, usaged in all_users_usaged.most_common():
            usaged = usaged / ((ls[-1][0] - ls[0][0]).total_seconds() / 3600 + 1e-9)
            chart.add_series(f'{user}{pretty_val(usaged):g}',
                             to_list([x[2].get(user, 0) for x in ls]),
                             areaStyle={}, stack='Total', emphasis={'focus': 'series'})

        return get_render_body(chart), sum(all_users_usaged.values())

    def dbview_cpu(self, recent=datetime.timedelta(days=1), date_trunc='hour'):
        from pyxllib.data.echarts import render_echart_html

        args = ['CPU核心数（比如4核显示是400%）', date_trunc, recent, 'sum(hosts.cpu_number)*100']

        htmltexts = []
        res = self._get_host_trace_total('cpu', 'XLPR服务器 CPU 使用近况', *args)
        htmltexts.append(res[0])

        hosts = self.execute('SELECT host_name, nick_name FROM hosts WHERE gpu_gb > 0').fetchall()
        host_stats = []
        for i, (hn, nick_name) in enumerate(hosts, start=1):
            name = f'{hn}，{nick_name}' if nick_name else hn
            host_stats.append(self._get_host_trace_per_host(hn, 'cpu', f'{name}', *args))
        host_stats.sort(key=lambda x: -x[1])  # 按使用量，从多到少排序
        htmltexts += [x[0] for x in host_stats]

        self.commit()

        h = render_echart_html('cpu', body='<br/>'.join(htmltexts))
        return h

    def dbview_cpu_memory(self, recent=datetime.timedelta(days=1), date_trunc='hour'):
        from pyxllib.data.echarts import render_echart_html

        args = ['内存（单位：GB）', date_trunc, recent, 'sum(hosts.cpu_gb)']

        htmltexts = []
        res = self._get_host_trace_total('cpu_memory', 'XLPR服务器 内存 使用近况', *args)
        htmltexts.append(res[0])

        hosts = self.execute('SELECT host_name, nick_name FROM hosts WHERE gpu_gb > 0').fetchall()
        host_stats = []
        for i, (hn, nick_name) in enumerate(hosts, start=1):
            name = f'{hn}，{nick_name}' if nick_name else hn
            host_stats.append(self._get_host_trace_per_host(hn, 'cpu_memory', f'{name}', *args))
        host_stats.sort(key=lambda x: -x[1])  # 按使用量，从多到少排序
        htmltexts += [x[0] for x in host_stats]

        self.commit()

        h = render_echart_html('cpu_memory', body='<br/>'.join(htmltexts))
        return h

    def dbview_disk_memory(self, recent=datetime.timedelta(days=30), date_trunc='day'):
        """ 查看disk硬盘使用近况
        """
        from pyxllib.data.echarts import render_echart_html

        args = ['硬盘（单位：GB）', date_trunc, recent, 'sum(hosts.disk_gb)']

        htmltexts = []
        res = self._get_host_trace_total('disk_memory', 'XLPR服务器 DISK硬盘 使用近况', *args)
        htmltexts.append(res[0])
        htmltexts.append('注：xlpr4（四卡）服务器使用du计算/home大小有问题，未统计在列<br/>')

        hosts = self.execute('SELECT host_name, nick_name FROM hosts WHERE gpu_gb > 0').fetchall()
        host_stats = []
        for i, (hn, nick_name) in enumerate(hosts, start=1):
            name = f'{hn}，{nick_name}' if nick_name else hn
            host_stats.append(self._get_host_trace_per_host(hn, 'disk_memory', f'{name}', *args))
        host_stats.sort(key=lambda x: -x[1])  # 按使用量，从多到少排序
        htmltexts += [x[0] for x in host_stats]

        self.commit()

        h = render_echart_html('disk_memory', body='<br/>'.join(htmltexts))
        return h

    def dbview_gpu_memory(self, recent=datetime.timedelta(days=30), date_trunc='day'):
        """ 查看gpu近使用近况

        TODO 这里有可以并行处理的地方，但这个方法不是很重要，不需要特地去做加速占用cpu资源
        """
        from pyxllib.data.echarts import render_echart_html

        args = ['显存（单位：GB）', date_trunc, recent, 'sum(hosts.gpu_gb)']

        htmltexts = []
        res = self._get_host_trace_total('gpu_memory', 'XLPR八台服务器 GPU显存 使用近况', *args)
        htmltexts.append(res[0])

        hosts = self.execute('SELECT host_name, nick_name FROM hosts WHERE gpu_gb > 0').fetchall()
        host_stats = []
        for i, (hn, nick_name) in enumerate(hosts, start=1):
            name = f'{hn}，{nick_name}' if nick_name else hn
            host_stats.append(self._get_host_trace_per_host(hn, 'gpu_memory', f'{name}', *args))
        host_stats.sort(key=lambda x: -x[1])  # 按使用量，从多到少排序
        htmltexts += [x[0] for x in host_stats]

        self.commit()

        h = render_echart_html('gpu_memory', body='<br/>'.join(htmltexts))
        return h

    def __4_一些数据更新操作(self):
        """ 比如一些扩展字段，在调用api的时候为了性能并没有进行计算，则可以这里补充更新 """

    def update_files_dhash(self, print_mode=True):
        """ 更新files表中的dhash字段值 """
        from pyxllib.cv.imhash import dhash
        from pyxllib.xlcv import xlpil

        # 获取总图片数
        total_count = self.execute("SELECT COUNT(*) FROM files WHERE dhash IS NULL").fetchall()[0][0]

        # 执行查询语句，获取dhash为NULL的记录

        # 初始化进度条
        progress_bar = tqdm(total=total_count, disable=not print_mode)

        while True:
            result = self.execute("SELECT id, data FROM files WHERE dhash IS NULL LIMIT 5")
            if not result:
                break
            for row in result:
                file_id = row[0]
                im = xlpil.read_from_buffer(row[1])
                computed_dhash = str(dhash(im))
                self.update_row('files', {'dhash': computed_dhash}, {'id': file_id})
                progress_bar.update(1)
            self.commit()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2023/07/13 14:26

from pyxllib.prog.pupil import check_install_package

check_install_package('transformers', 'transformers')

import ast
from collections import OrderedDict
from collections import Counter
import contextlib
import copy
import datetime
import heapq
import html
import json
import math
import random
import re
from urllib.parse import unquote
import io
import logging
import warnings

from jinja2 import Template
from openpyxl import Workbook
import pandas as pd
import requests
from tqdm import tqdm

try:
    from transformers import GPT2TokenizerFast
except ModuleNotFoundError:
    pass

from pyxllib.prog.pupil import OutputLogger
from pyxllib.prog.specialist import browser, TicToc
from pyxllib.algo.pupil import ValuesStat
from pyxllib.file.specialist import XlPath, JsonlDataFile, JsonlDataDir, TwinDirs, ensure_localdir
from pyxllib.file.xlsxlib import extract_workbook_summary


def __1_生成提问数据():
    pass


class Tokenizer:
    _tokenizer = None

    @classmethod
    def get_tokenizer(cls):
        """ 获取tokenizer，第一次调用时进行初始化 """

        if Tokenizer._tokenizer is None:
            # 根本没必要每次都尝试连接官网，本地有就不要老是sb的尝试连接huggingface
            # 而且官网连接也不稳，这里换成我自己的服务器中转
            gpt2_dir = XlPath.tempdir() / 'huggingface_gpt2'
            ensure_localdir(gpt2_dir, 'https://xmutpriu.com/download/huggingface_gpt2.zip')
            Tokenizer._tokenizer = GPT2TokenizerFast.from_pretrained(gpt2_dir)
        return Tokenizer._tokenizer

    @classmethod
    def tokenize(cls, paragraph, max_length=500):
        """ 对段落进行tokenize

        :param str paragraph: 待分词的段落
        :param int max_length: 单次处理的最大分词数，为了防止超过GPT2的限制，默认设置为500
        :return list: 分词后的列表

        >>> Tokenizer.tokenize('Hello, world! 汉字 123.14 35')
        ['Hello', ',', 'Ġworld', '!', 'Ġæ', '±', 'ī', 'åŃ', 'Ĺ', 'Ġ123', '.', '14', 'Ġ35']
        """
        tokenizer = Tokenizer.get_tokenizer()

        # 对段落进行切分
        paragraph_slices = [paragraph[i:i + max_length] for i in range(0, len(paragraph), max_length)]

        # 对每个切分的子段进行分词，并将结果拼接在一起
        tokens = []
        for slice in paragraph_slices:
            tokens += tokenizer.tokenize(slice)

        return tokens

    @classmethod
    def count_tokens(cls, paragraph, max_length=500):
        """ 获取段落的token数量

        :param str paragraph: 待分词的段落
        :param int max_length: 单次处理的最大分词数，为了防止超过GPT2的限制，默认设置为500
        :return int: token的数量

        >>> Tokenizer.count_tokens('Hello, world!')
        5
        """
        return len(cls.tokenize(paragraph, max_length))


def print_statistics(data, indent_level=1):
    """ 计算字符串长度，并且计算关键的一些token数

    :param data: data应该是一个嵌套结构，表示会话与消息
    """
    fmts = ['g', '.0f', '.0f', 'd', 'd']
    stat_len = ValuesStat([len(str(x)) for x in data])

    indent = '\t' * indent_level
    print(f'{indent} {stat_len.summary(fmts)}')


def check_conversation_lengths(all_texts, n_values=(4, 4),
                               compute_tokens=False, ids=None):
    """ 分析会话长度 """

    # 0 预处理
    for i, texts in enumerate(all_texts):
        if isinstance(texts, str):
            all_texts[i] = [texts]

    # 如果没有提供ID，则使用默认的range(n)
    if ids is None:
        ids = list(range(len(all_texts)))

    # 处理n_values的重叠
    if sum(n_values) >= len(all_texts):
        n_values = [len(all_texts), 0]  # 将所有数据视为最短数据，不再考虑最长数据

    # 1 消息长度统计
    fmts = [None, '.0f', '.0f', 'd', 'd']
    lengths = [len(t) for texts in all_texts for t in texts]
    print(f'1、消息长度统计 {ValuesStat(lengths).summary(fmts)}')

    # 2 每组会话消息数目
    ct = Counter(len(texts) for texts in all_texts)
    sorted_ct = {k: v for k, v in sorted(ct.items(), key=lambda x: x[0])}
    print(f'2、每组消息数目: {sorted_ct}')

    # 3 找出消息总长度最短和最长的会话
    total_lengths = [(i, sum(len(t) for t in texts)) for i, texts in enumerate(all_texts)]
    shortest_indices = [item[0] for item in heapq.nsmallest(n_values[0], total_lengths, key=lambda x: x[1])]
    longest_indices = [item[0] for item in heapq.nlargest(n_values[1], total_lengths, key=lambda x: x[1])]
    longest_indices = longest_indices[::-1]  # 从小到大排序

    parts = []
    if shortest_indices:
        parts.append(', '.join(map(str, [ids[i] for i in shortest_indices])))
    if longest_indices:
        parts.append(', '.join(map(str, [ids[i] for i in longest_indices])))
    print(f'3、最短最长会话的id:', ', ..., '.join(parts))

    # 4 计算token
    if compute_tokens:
        # 4.1 代表性样本的tokens数
        s_texts = [' '.join([x for x in all_texts[i]]) for i in shortest_indices]
        l_texts = [' '.join([x for x in all_texts[i]]) for i in longest_indices]

        s_lens = [[len(x), Tokenizer.count_tokens(x)] for x in s_texts]
        l_lens = [[len(x), Tokenizer.count_tokens(x)] for x in l_texts]

        # try:
        #     s_lens = [[len(x), Tokenizer.count_tokens(x)] for x in s_texts]
        #     l_lens = [[len(x), Tokenizer.count_tokens(x)] for x in l_texts]
        # except Exception as e:
        #     # 如果token相关功能用不了，则不用继续处理，结束该函数即可
        #     return

        parts = []
        if s_lens:
            parts.append(', '.join(map(str, [x[1] for x in s_lens])))
        if l_lens:
            parts.append(', '.join(map(str, [x[1] for x in l_lens])))
        # 仅计算3中代表性样本
        print(f'4、tokens数量:', ', ..., '.join(parts))

        # 4.2 token的比率规律
        ratios = []
        for x in s_lens + l_lens:
            ratios.append(x[1] / x[0])
        fmts = [None, '.0%', '.0%', '.0%', '.0%']
        print(f'token/len比率统计 {ValuesStat(ratios).summary(fmts)}')
        # 比率越大，代表越接近中文场景，汉字越多，要注意len的控制不要让token某些场合超出长度


def set_template(s, *args, **kwargs):
    """ todo 这个名字会不会太容易冲突了？ """
    return Template(s.strip(), *args, **kwargs)


def set_meta_template(s, meta_start='[[', meta_end=']]', **kwargs):
    """ 支持预先用某些格式渲染后，再返回标准渲染模板 """
    t = Template(s.strip(), variable_start_string=meta_start,
                 variable_end_string=meta_end).render(**kwargs)
    return Template(t)


class GptChatJsonl(JsonlDataFile):
    """ GPT问答批量执行脚本的jsonl生成、读取器 """

    def __init__(self, file=None, num_records=None, *, start_id=None):
        from datetime import datetime

        super().__init__(file, num_records)
        if start_id is None:
            # 230821周一02:02，原本只有日期标记，后面发现这样id很容易出现重复，还是加上小时分钟更不容易引起一些没必要的麻烦
            today = datetime.now().strftime("%Y%m%d%H%M")
            self.start_id = int(today + "000000")
        else:
            self.start_id = start_id

    def read_jsonl(self, file):
        """ 从一个文件加载数据
        """
        self.records = XlPath(file).read_jsonl()
        try:
            self.start_id = self.records[-1]['id']
        except KeyError:
            pass

    def split_and_add_prompt(self, text, max_word_length=None, prompt=None):
        """
        :param text: 要插入的文本（纯文本，不能是字典格式的 {'content': ...}）
        :param max_word_length: 如果设置了该值，则会对输入内容进行长度控制，拆成多个片段
            之前测试，全英文长度大概在32500，中文在5000内
        :param prompt: max_word_length开启时才会生效，每个part要附加的提示规则
            可以写个字符串，默认每段都是开头加这个提示
            也可以参考gen_prompt2,一些生成函数写法进行自定义
        :return:
        """
        # 0 如果没有输入max_word_length，就不用特地处理了
        if max_word_length is None:
            return [text]

        # 1 工具函数
        def gen_prompt1(n, i, text):
            """ 一共n条，当前第i条的，当前内容是text """
            if n == 1:
                return text
            if n > 1:
                if i == 0:
                    return f'【注意，由于本次提问过长，这里拆分成{n}个片段分开输入，目前是第{1}个片段，你只需暂时回复"收到"即可】\n' + text
                elif i < n - 1:
                    return f'【这是{n}个片段中的第{i + 1}个片段，先回复"收到"即可】\n' + text
                else:
                    return f'【这是{n}个片段的最后一个片段，请开始回复内容】\n' + text

        def gen_prompt2(n, i, text):
            return prompt + text

        if prompt is None:
            gen_prompt = gen_prompt1
        elif isinstance(prompt, str):
            gen_prompt = gen_prompt2
        else:  # callable
            gen_prompt = prompt

        # 2 拆分重拼接
        # 首先要重新调整max_word_length，在确定要拆分为几个片段的情况下，尽量保证这些片段之间的均匀性
        num = len(text) // max_word_length + 1
        max_word_length = math.ceil(len(text) / num)

        # 2.1 检查是否有超长单行文本，要提前切分成多行
        lines = text.rstrip().split('\n')
        new_lines = []
        for line in lines:
            if len(line) < max_word_length:
                new_lines.append(line)
            else:  # 单行就已经爆限制长度的比较特别
                n = max_word_length - 10  # 将这个长文本按照n的长度再拆分成多个片段加入new_lines
                parts = [line[i:i + n] for i in range(0, len(line), n)]
                new_lines += parts

        # 2.2 拼接new_lines
        fragments = []
        current_fragment = []

        for line in new_lines:
            if len(current_fragment) + len(line) <= max_word_length:
                current_fragment.append(line)
            else:
                fragments.append('\n'.join(current_fragment))
                current_fragment = [line]
        if current_fragment:
            fragments.append('\n'.join(current_fragment))

        n = len(fragments)
        fragments = [gen_prompt(n, i, x).strip() for i, x in enumerate(fragments)]
        return fragments

    def split_texts(self, texts, max_word_length=None, prompt=None):
        """ 长对话自动拆分成多轮对话 """
        new_texts = []
        for text in texts:
            pure_text = text['content']
            new_texts += self.split_and_add_prompt(pure_text, max_word_length=max_word_length, prompt=prompt)
            if 'file_paths' in text:  # 如果有文件，自动放在最后一轮插入
                new_texts[-1]['file_paths'] = text['file_paths']
        return new_texts

    def add_record(self, texts, *, extra=None,
                   record_id=0, max_word_length=None, prompt=None):
        """
        :param texts:
            str -> list[str]，可以只输入一个str，默认一轮对话
            list[str] -> list[{'content': ..., 'file_paths': [...]}]
                content: 文本内容
                file_paths: 注意可以设置本地电脑其他来源，会自动移到该任务的upload_files里
        :param record_id: 可以自定义这个session的id
        :param max_word_length: 是否设置一个约束长度，自动切分会话中太长的消息
            gpt4是8192个token，大概len就是8192/0.6=13653，一般建议如果要设就设10000左右
        :param prompt: 自动分段后
            None，自动配置的一套提示
            '', 不用提示
        :return:
        """
        # 1 变成标准的list + 字典结构，方便后面统一处理
        if not isinstance(texts, list):
            texts = [texts]

        for i, text in enumerate(texts):
            if isinstance(text, str):
                texts[i] = {'content': text}

        # 2 如果设置了每次最大会话长度，要进行拆分
        if max_word_length:
            texts = self.split_texts(texts, max_word_length=max_word_length, prompt=prompt)

        for i, text in enumerate(texts):
            texts[i]['content'] = text['content'].strip()

        # 3 添加会话conversation
        self.start_id += 1
        item = {'id': str(record_id or self.start_id),  # 要转成字符串类型，不然容易出问题
                'text': texts,
                'first_text_length': len(texts[0]['content'])}
        if extra:
            item['extra'] = extra
        self.records.append(item)
        return item

    def fix_file_paths(self, save_dir):
        """ 修正records中设置的file_paths

        这些路径可能在设置的时候图方便，设置的是非项目目录下的路径
        这个函数会对这些路径进行修正，为了修正，需要输入一个该jsonl所保存的目录位置
        """
        save_dir = XlPath(save_dir)
        for i, record in tqdm(enumerate(self.records), desc='修复文件路径'):
            dst_dir = save_dir / 'upload_files' / str(record['id'])
            for j, text in enumerate(record['text']):
                for k, fn in enumerate(text.get('file_paths', [])):
                    src_file = XlPath(fn)
                    src_file2 = src_file.as_posix()
                    if src_file2.startswith(f'upload_files/{record["id"]}/'):
                        continue
                    dst_file = dst_dir / src_file.name
                    dst_file2 = dst_file.relpath(save_dir).as_posix()
                    if src_file.is_file():
                        if src_file2 != dst_file2:
                            dst_dir.mkdir(parents=True, exist_ok=True)
                            src_file.copy(dst_file, if_exists='replace')
                    else:  # 既然设置了，原文件目录应该在
                        raise FileNotFoundError(f'{src_file}')
                    text['file_paths'][k] = dst_file2

    def clean_file_paths(self):
        """ 清除records中的file_paths
        一般用于把一些相关文件移到对应会话后，实际提问gpt的时候并不上传文件
        """
        for x in self.records:
            for t in x['text']:
                if 'file_paths' in t:
                    del t['file_paths']

    def find_indices_by_qlength(self):
        """ 返回提问(q,question)内容从短到长的数据下标 """
        lens = [(i, len(''.join([t['content'] for t in x['text']]))) for i, x in enumerate(self.records)]
        # 根据长度进行排序，得到的元组的第一个元素为原列表的下标，第二个元素为对应的长度
        sorted_lens = sorted(lens, key=lambda x: x[1])
        # 取出排序后的下标
        sorted_indexs = [i for i, _ in sorted_lens]
        return sorted_indexs

    def browse_record(self, index=None, paths=None, **kwargs):
        """ 检查第i次会话的内容
        """
        # 如果未提供索引，则尝试使用查询参数找到第一个匹配的记录
        if index is None:
            index = self.find_index(paths, **kwargs)
            if index is None:
                raise ValueError('No matching record found')
        session = self.records[index]

        # 构建HTML内容
        html_content = "<html><body>"

        # 输出除了text和all_answers以外的所有键值信息
        html_content += "<h2>会话信息：</h2>"
        html_content += "<ul>"
        for key, value in session.items():
            if key not in ["text", "all_answers"]:
                html_content += f"<li>{html.escape(key)}: {html.escape(str(value))}</li>"
        html_content += "</ul>"

        # 输出text和all_answers的内容
        texts = self.get_text_texts(session.get("text", []))
        all_answers = self.get_all_answers_texts(session.get("all_answers", []))

        max_length = max(len(texts), len(all_answers))
        for idx in range(max_length):
            html_content += f"<h3>第{idx + 1}次询问：</h3>"
            if idx < len(texts):
                html_content += f"<pre>{html.escape(texts[idx])}</pre>"
            if idx < len(all_answers):
                html_content += f"<h3>第{idx + 1}次回答：</h3>"
                html_content += f"<pre>{html.escape(str(all_answers[idx]))}</pre>"

        html_content += "</body></html>"
        html_file = (XlPath.tempdir() / (str(session.get('id', index)) + '.html'))
        html_file.write_text(html_content)
        browser.html(html_file)

        # 返回HTML字符串
        return html_content

    def get_text_texts(self, text):
        """ 从text字段获得所有的文本内容
        因为里面可能是dict
        """
        ls = []
        for t in text:
            if isinstance(t, str):
                ls.append(t)
            else:
                if "file_path" in t:
                    ls.append(("filep_path=" + str(t["file_path"]) + "\n\n") + t["content"])
                else:
                    ls.append(t["content"])
        return ls

    def get_all_answers_texts(self, all_answers):
        ls = []
        for t in all_answers:
            if isinstance(t, dict):
                t = json.dumps(t, ensure_ascii=False, indent=2)
            ls.append(str(t))
        return ls

    def check(self):
        """ 检查会话、消息长度等信息 """
        # 1 提问的内容
        all_texts = [self.get_text_texts(session.get('text', []))
                     for session in self.records]
        print('【提问的内容】')
        check_conversation_lengths(all_texts,
                                   compute_tokens=True,
                                   ids=[x['id'] for x in self.records])

        # 2 回复的内容
        all_texts = [self.get_all_answers_texts(session.get('all_answers', []))
                     for session in self.records]
        # 过滤空值，并相应地更新ids
        filtered_texts = [(text, session['id']) for text, session in zip(all_texts, self.records) if text]
        all_texts, ids = zip(*filtered_texts) if filtered_texts else ([], [])
        if all_texts:
            print('【回复的内容】')
            check_conversation_lengths(all_texts,
                                       compute_tokens=True,
                                       ids=ids)

    def filter_records_without_answers(self):
        """ 过滤掉没有 'all_answers' 字段的sessions """

        # 输出过滤前的sessions数量
        print(f"过滤前的sessions数量：{len(self.records)}")

        # 使用列表推导式过滤出包含 'all_answers' 字段的sessions
        self.records = [s for s in self.records
                        if (''.join(map(str, s.get('all_answers', []))))]

        # 输出过滤后的sessions数量
        print(f"过滤后的sessions数量：{len(self.records)}")

    @classmethod
    def _parse_single_record_answer_contents(cls, record):
        """ 注意本函数不做record备份 """
        for answer in record.get('all_answers', []):
            if isinstance(answer, dict) and 'contents' in answer:
                n = len(answer['contents'])
                for i in range(n - 1, -1, -1):
                    message = answer['contents'][i]['message']
                    if message and 'content' in message and 'error' not in message:
                        break
                else:
                    answer['contents'] = ''
                    continue

                content = message['content']
                if 'parts' in content:
                    content = '\n'.join(content['parts'])
                else:
                    content = content['text']
                answer['contents'] = content

    @classmethod
    def _parse_single_record_answer_downloads(cls, record):
        for answer in record.get('all_answers', []):
            if 'downloads' in answer:
                for i, link in enumerate(answer['downloads']):
                    m = re.search(r'filename%3D(.+?)&sig=', link)
                    if m:
                        answer['downloads'][i] = unquote(unquote(m.group(1)))

    @classmethod
    def parse_single_record_answer(cls, record):
        cls._parse_single_record_answer_contents(record)
        cls._parse_single_record_answer_downloads(record)

    def parse_answer_contents(self):
        """ 简化解释器返回结果中，contents的结构信息 """
        for record in self.records:
            self._parse_single_record_answer_contents(record)

    def parse_answer_downloads(self):
        """ 解析，简化下载链接的表达形式 """
        for record in self.records:
            self._parse_single_record_answer_downloads(record)

        # 目录里的文件名也同理做精简
        for f in self.infile.parent.glob_files():
            if f.name.startswith('OpenAI-download-'):
                f.rename2(f.parent / re.sub(r'OpenAI-download-\d+-', '', f.name),
                          if_exists='replace')

    def filter_to_rechat(self, check_func, rechat_path=None):
        """ 筛选失败的数据到一个新的目录，常用于对chatted数据筛选出未成功的样例，上池子重跑
        这个不是简单的找出得不到all_answers的，而是可以很精细，包含复杂post、verify的情况

        :param check_func: 一个函数，接收一个record，返回True或False
            True，表示这个record是对的
            False，表示这个record是错的，要挑选出来
        :param rechat_path: 把挑选出来的数据放到新路径
        """
        if rechat_path is None:
            rechat_path = XlPath(self.infile.parent.as_posix() + '_rechat/in.jsonl')

        rechat_path = XlPath(rechat_path)
        td = TwinDirs(self.infile.parent, rechat_path.parent)

        gcj = type(self)()
        for record in self.records:
            if not check_func(record):
                record2 = {}
                for k in ['id', 'text', 'first_text_length', 'extra']:
                    record2[k] = record[k]
                gcj.records.append(record2)
                for x in record['text']:
                    if 'file_path' in x:
                        td.copy_file(td.src_dir / x['file_path'])

        gcj.save(rechat_path)
        return gcj

    def update_from_rechat(self, check_func, rechat_path=None):
        """ 从另一份rechat的数据，更新回主master数据

        :param check_func: 原chatted没过，但是rechatted通过的，需要把数据更新过来
        :param rechat_path: 注意只能传路径，因为可能涉及到文件操作，需要知道目录所在
            依据这个文件里的record记录更新回self
        """
        if rechat_path is None:
            rechat_path = XlPath(self.infile.parent.as_posix() + '_rechat') / 'out.jsonl'

        rechat_path = XlPath(rechat_path)
        td = TwinDirs(rechat_path.parent, self.infile.parent)

        id2index = {x['id']: i for i, x in enumerate(self.records)}

        gcj = type(self)(rechat_path)
        gcj.parse_answer_contents()
        gcj.parse_answer_downloads()

        # 需要处理下下载链接名称
        self.parse_answer_downloads()
        gcj.parse_answer_downloads()

        for y in gcj.records:
            index = id2index[y['id']]
            x = self.records[index]
            if not check_func(x) and check_func(y):
                # 先把x相关的数据删掉
                if 'all_answers' in x:
                    for answer in x['all_answers']:
                        for fname in answer.get('downloads', []):
                            (XlPath(self.infile.parent) / fname).delete()
                # 再把y拷贝过来
                for answer in y['all_answers']:
                    for fname in answer.get('downloads', []):
                        td.copy_file(td.src_dir / fname)
                self.records[index] = y
        return gcj


GptQuestionJsonl = GptChatJsonl  # 名称向下兼容


def __2_数据后处理():
    """ 一些常用的文本、后处理功能也放到这里 """


def try_eval_json(resp_json):
    try:
        resp_json = ast.literal_eval(resp_json)
        if isinstance(resp_json, dict):
            resp_json = resp_json[resp_json.keys()[0]]
    except:
        pass
    return resp_json


def try_load_json(resp_json):
    if isinstance(resp_json, str):
        try:
            resp_json = json.loads(resp_json)
            if isinstance(resp_json, dict):
                resp_json = resp_json[resp_json.keys()[0]]
        except:
            pass
    return resp_json


def try_parse_json(resp_json):
    if isinstance(resp_json, dict):
        try:
            resp_json = '\n'.join(resp_json['contents'][-1]['message']['content'].get('parts', []))
        except TypeError:
            return ''

    resp_json = try_eval_json(resp_json)
    if isinstance(resp_json, str):
        return try_load_json(resp_json)
    return resp_json


def extract_code_blocks_from_md(markdown_text, *, sort_by_length=False):
    """ 可以输入str，也可以输入list[str]

    :param sort_by_length: 按代码长度从短到长排序
        常用在比较确信有效代码段应该只有一段，但是有些短小的片段有干扰
        此时可以排序后，选取最长的一个代码片段作为正确代码
    """
    if isinstance(markdown_text, str):
        markdown_text = [markdown_text]

    matches = []
    pattern = re.compile(r'^```[^\n]*\n(.+?)\n^```', re.MULTILINE | re.DOTALL)
    for text in markdown_text:
        matches += pattern.findall(text)

    if sort_by_length:
        matches = sorted(matches, key=len)

    return matches


def extract_airscript_code_from_answers(all_answers):
    """ 从多轮回答的最后一次回答中提取求解代码 """
    contents = all_answers[-1]['contents']
    text = contents[-1]['text']
    code_blocks = extract_code_blocks_from_md(text, sort_by_length=True)

    if code_blocks:
        return code_blocks[-1]
    else:
        return ''


def __3_生成最后训练用的数据():
    pass


def texts2train_record(texts):
    """ user和assistant的轮询对话，转为训练集格式 """
    messages = []
    for i, text in enumerate(texts):
        role = 'assistant' if i % 2 else 'user'
        messages.append({'role': role, 'content': text})
    return {'messages': messages}


class GptTrainJsonl(JsonlDataFile):
    """
    record: dict
        messages: list
          dict: role='user', content=...
          dict: role='assistant', content=...
    """

    def analyze_text_length(self):
        # 1 先将数据统计到df
        ls = []
        columns = ['role', 'content']
        for x in self.records:
            for t in x['messages']:
                ls.append([t['role'], t['content']])
        df = pd.DataFrame.from_records(ls, columns=columns)

        # 2 再从df筛选出不同的统计数据
        print('【user和assistant】')
        print_statistics(df['content'])
        print('【user】')
        print_statistics(df[df['role'] == 'user']['content'])
        print('【assistant】')
        print_statistics(df[df['role'] == 'assistant']['content'])

    def check(self):
        """ 检查会话、消息长度等信息 """
        # 1. 提取'user'角色的content
        user_texts = [[message['content']
                       for message in record['messages']
                       if message['role'] == 'user']
                      for record in self.records]
        print('【User的内容】')
        check_conversation_lengths(user_texts, compute_tokens=True,
                                   ids=list(range(len(user_texts))))

        # 2. 提取'assistant'角色的content
        assistant_texts = [[message['content']
                            for message in record['messages']
                            if message['role'] == 'assistant']
                           for record in self.records]
        print('【Assistant的内容】')
        check_conversation_lengths(assistant_texts, compute_tokens=True,
                                   ids=list(range(len(assistant_texts))))

        # 3. 将整个record视为一个完整的会话
        full_conversations = [' '.join([message['content'] for message in record['messages']])
                              for record in self.records]
        print('【完整的会话】')
        check_conversation_lengths(full_conversations, compute_tokens=True,
                                   ids=list(range(len(full_conversations))))

    def browse_record(self, index=None, paths=None, **kwargs):
        """ 显示第i次会话的内容 """
        # 如果未提供索引，则尝试使用查询参数找到第一个匹配的记录
        if index is None:
            index = self.find_index(paths, **kwargs)
            if index is None:
                raise ValueError('No matching record found')
        session = self.records[index]

        # 构建HTML内容
        html_content = "<html><body>"

        # 输出除了messages以外的所有键值信息
        html_content += "<h2>会话信息：</h2>"
        html_content += "<ul>"
        for key, value in session.items():
            if key != "messages":
                html_content += f"<li>{html.escape(key)}: {html.escape(str(value))}</li>"
        html_content += "</ul>"

        # 输出messages的内容
        messages = session.get("messages", [])

        for idx, message in enumerate(messages):
            role = message.get('role', 'unknown')
            content = message.get('content', '')
            html_content += f"<h3>第{(idx // 2) + 1}次{role}的发言：</h3>"
            html_content += f"<pre>{html.escape(content)}</pre>"

        html_content += "</body></html>"
        html_file = (XlPath.tempdir() / (f'session_{index}.html'))  # 创建临时文件名，防止覆盖现有文件
        html_file.write_text(html_content)
        browser.html(html_file)  # 在浏览器中打开HTML文件

        # 或者返回HTML字符串
        return html_content

    def add_record(self, texts):
        messages = []
        for i, text in enumerate(texts):
            role = 'assistant' if i % 2 else 'user'
            messages.append({'role': role, 'content': text})
        self.records.append({'messages': messages})

    def add_from_texts(self, texts):
        record = texts2train_record(texts)
        self.records.append(record)


def __4_综合集成类():
    pass


class GptChatDir:
    """ 一个目录，包含了一个任务的所有数据，包括in、out、post等文件 """

    def __init__(self, root=None, lines_per_file=10000):
        if root is None:
            root = self.__class__.__name__.lower()

        self.root = root = XlPath(root)
        self.lines_per_file = lines_per_file

        self.chat_file = root / 'in.jsonl'
        self.chatted_file = root / 'out.jsonl'
        self.post_file = root / 'post.jsonl'
        self.verify_file = root / 'verify.jsonl'
        self.train_file = root / 'train.jsonl'

        # 如果有目录文件，会优先以目录为准。如果没有，则会从单文件拆分创建。
        self.update_dir()

        self.upload_files_dir = root / 'upload_files'
        self.download_files_dir = root / 'download_files'

        # todo 把 1chat 改名 in，2chatted 改名 out
        # for f in self.root.glob_files('*1chat*.jsonl'):
        #     f.rename2(f.parent / 'in.jsonl')

        # for dir_path in [self.root, self.upload_files_dir, self.download_files_dir]:
        for dir_path in [self.root]:
            if not dir_path.is_dir():
                dir_path.mkdir(parents=True, exist_ok=True)

        # 这个类经常要并发处理，不能把一个不能序列化的类放到这里~
        # self.logger = OutputLogger(log_file=self.root / 'log.txt')

    def update_dir(self):
        """ 目录结构有些更新后，一些成员变量要跟着改变 """
        # 如果有目录文件，会优先以目录为准。如果没有，则会从单文件拆分创建。
        self.chat_dir = JsonlDataDir.init_from_file(self.chat_file, self.lines_per_file)
        self.chatted_dir = JsonlDataDir.init_from_file(self.chatted_file, self.lines_per_file)
        self.post_dir = JsonlDataDir.init_from_file(self.post_file, self.lines_per_file)
        self.verify_dir = JsonlDataDir.init_from_file(self.verify_file, self.lines_per_file)
        self.train_dir = JsonlDataDir.init_from_file(self.train_file, self.lines_per_file)

    def summary_records(self):
        """ 一些统计信息 """
        # 1 chat信息
        gcd1 = self.chatted_dir or self.chat_dir
        if not gcd1:
            print('请确认是否有生成初始的chat数据')
            return

        print(f'【{self.root.name}】')
        texts = [len(x['text']) for x in gcd1.yield_record()]
        n, m = len(texts), sum(texts)
        print(f'1、chat：{n}条会话*{m / n:.2g}条消息')
        gcj1 = GptChatJsonl(gcd1.files[0])  # 统计一个文件就够了，不然太多了
        gcj1.check_records()
        print()

        # 2 chatted信息
        filter_records = [x for x in gcd1.yield_record() if 'all_answers' in x]
        if filter_records:
            print(f'2、chatted：已获得{len(filter_records)}条会话')
        else:
            print('2、chatted：暂未获得生成数据')

        # 3 post信息
        if self.post_dir:
            print(f'3、post：{self.post_dir.count_records()}条会话')

        # 4 verify（这一步有时候会集成到post中）
        if self.verify_dir:
            print(f'4、verify：{self.verify_dir.count_records()}条会话')

        # 5 train 生成的训练数据
        # print('5、train：')
        # gtj = GptTrainJsonl(self.train_file)
        # gtj.analyze_text_length()

    def summary_downloads(self):
        """ 统计下载的文件情况 """
        print('【每个目录文件数量】')
        files_each_dir = []
        for d in self.download_files_dir.glob_dirs():
            files_each_dir.append(len(list(d.rglob_files())))
        print(ValuesStat(files_each_dir).summary())
        print(Counter(files_each_dir))

        print('【每个文件大小】')
        filesizes_each_dir = []
        for d in self.download_files_dir.glob_dirs():
            for f in d.rglob_files():
                filesizes_each_dir.append(f.size())
        print(ValuesStat(filesizes_each_dir).summary())

    def create_chat(self):
        """ 生成chat数据，具体内容方式跟业务有关 """
        raise NotImplementedError

    def browse_chatted_record(self, index=None, paths=None, **kwargs):
        """ 显示第i次会话的内容 """
        f = self.chatted_file if self.chatted_file.is_file() else self.chat_file
        return GptChatJsonl(f, 100).browse_record(index, paths, **kwargs)

    def chatted2post_record(self, chatted_record):
        """ 后处理，解析

        一般会保留基本的all_answers结果，供检查上游一些基本情况
        然后把一些结构化结果存储到extra字段

        :return: 会返回新的dict结构的一个post_record，如果解析失败，会返回None
        """
        # 0 基本情况判断
        if 'all_answers' not in chatted_record:
            return

        post_record = copy.deepcopy(chatted_record)

        # 1 删掉一些没卵用的字段
        for name in ['all_questions', 'first_text_length', 'second_text_length']:
            if name in post_record:
                del post_record[name]

        # 2 解析all_answers：这个结构太复杂，进行内容整理精简
        # 2.1 contents：这个结构太复杂，搁这俄罗斯套娃呢~ 稍微精简下更方便后处理
        for k, answer in enumerate(post_record['all_answers']):
            if isinstance(answer, dict) and 'contents' in answer:
                new_contents = []
                for i, x in enumerate(answer['contents']):
                    if not x['message']:
                        # Error in message stream
                        # print(f'{post_record["id"]} answer[{k}] contents[{i}] message为空')
                        continue

                    content = x['message']['content']
                    tp = content['content_type']
                    new_content = {'type': content['content_type']}
                    if tp == 'text':
                        new_content['text'] = '\n'.join(content['parts'])
                    elif tp == 'code':
                        new_content['text'] = content['text']
                    elif tp == 'execution_output':
                        new_content['text'] = content['text']
                    elif tp == 'system_error':
                        continue
                    else:
                        print(f'{post_record["id"]} answer[{k}] contents[{i}] content_type={tp} 未见类型')
                        continue

                    new_contents.append(new_content)
                answer['contents'] = new_contents
            elif isinstance(answer, str):  # 普通模式也转成解释器风格，方便统一处理
                post_record['all_answers'][k] = {'contents': [{'type': 'text',
                                                               'text': answer}]}

        # 2.2 downloads：下载链接精简下，并把关联的文件也顺带整理一下
        for answer in post_record['all_answers']:
            if 'downloads' not in answer:
                continue
            for i, link in enumerate(answer['downloads']):
                m = re.search(r'filename%3D(.+?)&sig=', link)
                if m:
                    answer['downloads'][i] = str(post_record['id']) + '/' + unquote(unquote(m.group(1)))
                # 对应的文件不存在的不要，有数据超过50M的也不要
                file = self.download_files_dir / link
                if not file.exists() and file.size() > 50 * 1024 * 1024:
                    return

            # 理论上下载的文件不应该有重复，虽然不知道为什么会拿到重复，但去掉重复比较好
            answer['downloads'] = list(OrderedDict.fromkeys(answer['downloads']))

        # 2.3 删掉answer里其他没用的字段
        for answer in post_record['all_answers']:
            for name in ['created', 'message_id', 'conversation_id', 'end_turn']:
                if name in answer:
                    del answer[name]

        # 返回处理结果
        return post_record

    @staticmethod
    def post2verify_record(post_record):
        """ 这个一般是要具体任务定制的，没有通用操作方式

        注意，如果要使用create_verify的多进程功能，这个函数必须是静态的，否则写成类方法或对象方法都可以
        """
        raise NotImplementedError

    def verify2train_record(self, verify_record):
        """ 这个一般是要具体任务定制的，没有通用操作方式 """
        raise NotImplementedError

    def organize_downloaded_files(self):
        # 把下载的文件整理的更清晰些
        for f in tqdm(list(self.root.glob_files('OpenAI-download-*')),
                      desc='整理下载的文件'):
            new_name = re.sub(r'OpenAI-download-\d+-', '', f.name)
            new_name = new_name.replace('-', '/', 1)
            try:
                (self.download_files_dir / new_name).parent.mkdir(exist_ok=True)
                f.rename2(self.download_files_dir / new_name, if_exists='replace')
            except FileExistsError as e:
                # 有的文件会移动不了
                print(e)

        # 会剩一些特殊的处理不了的文件，可以看一眼后手动删掉
        # 这些相关的records，默认的chatted2post_record会把这些记录过滤掉

    def create_post(self, **kwargs):
        """ 建议初步跑的时候，先串行debug，等比较稳定后，再开并发跑
        """
        if 'dst_dir' not in kwargs:
            kwargs['dst_dir'] = self.post_dir.root
        self.chatted_dir.process_each_record(self.chatted2post_record, **kwargs)
        self.post_dir.update_subfiles()
        num1, num2 = self.chatted_dir.count_records(), self.post_dir.count_records()
        print(f'chatted有{num1}条，转换post有{num2}条，转换率{num2 / num1:.2%}')

    def create_verify(self, **kwargs):
        """ 有时候create_verify是有cpu密集运算场景的，可以开多进程
        """
        if 'dst_dir' not in kwargs:
            kwargs['dst_dir'] = self.verify_dir.root
        self.post_dir.process_each_record(self.post2verify_record, **kwargs)
        self.verify_dir.update_subfiles()
        num1, num2 = self.post_dir.count_records(), self.verify_dir.count_records()
        num1 = num1 or -1
        print(f'post有{num1}条，转换verify有{num2}条，转换率{num2 / num1:.2%}')

    def refine_verify(self, print_mode=1, **kwargs):
        """ 重复检查verify数据

        这个函数可以重复执行，但前提是self.post2verify_record里的设计有增量规则部分
        """
        self.verify_dir.process_each_record(self.post2verify_record, print_mode=print_mode,
                                            inplace=True, desc='refine_verify', **kwargs)

    @classmethod
    def texts2train_record(cls, texts):
        """ user和assistant的轮询对话，转为训练集格式 """
        messages = []
        for i, text in enumerate(texts):
            role = 'assistant' if i % 2 else 'user'
            messages.append({'role': role, 'content': text})
        return {'messages': messages}

    def create_train(self, **kwargs):
        if 'dst_dir' not in kwargs:
            kwargs['dst_dir'] = self.train_dir.root
        self.post_dir.process_each_record(self.verify2train_record, **kwargs)
        self.train_dir.update_subfiles()

    def check_chatted_record(self, chatted_record):
        """ 检查chatted数据的有效性 """
        x = chatted_record
        x = self.chatted2post_record(x)
        # x = self.post2verify_record(x)
        # 针对verify可以再进一步定制规则
        return bool(x)

    def create_rechat(self, rechat_path):
        """ 筛选失败的数据到一个新的目录，常用于对chatted数据筛选出未成功的样例，上池子重跑

        :param rechat_path: 把挑选出来的数据放到新路径
        """
        gcd = GptChatDir(rechat_path)
        f = open(gcd.chat_file, 'w', encoding='utf-8')

        for record in tqdm(self.chatted_dir.yield_record(), '检查待重新生成的问题'):
            if not self.check_chatted_record(record):
                continue
            # 否则把这个条目放到rechat，准备拿去重新提问
            if 'error' in record:
                del record['error']
            f.write(json.dumps(record, ensure_ascii=False) + '\n')
            # 如果有文件，也要对应移动
            src_dir = self.upload_files_dir / str(record['id'])
            if src_dir.is_dir():
                src_dir.copy(gcd.upload_files_dir / src_dir.name, if_exists='skip')

        f.close()
        return gcd

    def update_chatted(self, rechat_path):
        """ 从另一个rechat数据，更新数据条目过来

        self依然叫src，rechat叫dst，虽然其实数据是从rechat更新流向self

        注意：这个函数还没有比较严格地进行调试~
        """
        # 1 读取有效记录
        gcd = GptChatDir(rechat_path)
        gcd.organize_downloaded_files()
        # 请确保内存充足哦，这个函数会从rechat的chatted读取所有通过的记录保存起来
        dst_records = {}
        for record in gcd.chatted_dir.yield_record():
            # 找到有all_answers的挑出来
            post_record = self.chatted2post_record(record)
            if post_record:
                dst_records[record['id']] = record

        # 2 更新记录
        def update_each_record(x):
            if x['id'] in dst_records:
                # 除了返回record，还得拷贝目录数据呢
                # 上传的目录一般没变，但最好重置下
                src_dir = self.upload_files_dir / x['id']
                dst_dir = gcd.upload_files_dir / x['id']
                dst_dir.copy(src_dir, if_exists='replace')
                # 下载的目录
                src_dir = self.download_files_dir / x['id']
                dst_dir = gcd.download_files_dir / x['id']
                dst_dir.copy(src_dir, if_exists='replace')
                return dst_records[x['id']]
            else:
                return x

        self.chatted_dir.update_each_record(update_each_record)


def __5_bdchat():
    """ 百度相关api """


class BaiduChatbot:
    def __init__(self, api_key, secret_key, file_path=None):
        self.API_KEY = api_key
        self.SECRET_KEY = secret_key
        self.ACCESS_TOKEN = self._get_access_token()
        self.base_url = "https://aip.baidubce.com/rpc/2.0/ai_custom/v1/wenxinworkshop/chat/completions?access_token="
        self.file_path = file_path  # 文件路径为可选参数

    def _get_access_token(self):
        """
        使用 AK，SK 生成鉴权签名（Access Token）
        :return: access_token，或是None(如果错误)
        """
        url = "https://aip.baidubce.com/oauth/2.0/token"
        params = {
            "grant_type": "client_credentials",
            "client_id": self.API_KEY,
            "client_secret": self.SECRET_KEY
        }
        return str(requests.post(url, params=params).json().get("access_token"))

    def chat(self, user_message):
        """ 向Baidu API发送用户消息并返回API的回复
        注意user_message的token不要超过3k
        """
        url = self.base_url + self.ACCESS_TOKEN
        payload = json.dumps({
            "messages": [{"role": "user", "content": user_message}]
        })
        headers = {'Content-Type': 'application/json'}
        response = requests.post(url, headers=headers, data=payload)
        response_json = response.json()
        response_json['user_message'] = user_message
        response_json['timestamp'] = datetime.datetime.now().isoformat()

        # 如果指定了文件路径，自动保存记录
        if self.file_path:
            self._save_to_file(response_json)

        return response_json.get('result', '')

    def _save_to_file(self, response):
        with open(self.file_path, 'a', encoding='utf-8') as file:
            file.write(json.dumps(response, ensure_ascii=False) + '\n')



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import copy

__all__ = ["build_metric"]

from .det_metric import DetMetric
from .rec_metric import RecMetric
from .cls_metric import ClsMetric
from .e2e_metric import E2EMetric
from .distillation_metric import DistillationMetric
from .table_metric import TableMetric
from .kie_metric import KIEMetric


def build_metric(config):
    support_dict = [
        "DetMetric", "RecMetric", "ClsMetric", "E2EMetric",
        "DistillationMetric", "TableMetric", 'KIEMetric'
    ]

    config = copy.deepcopy(config)
    module_name = config.pop("name")
    assert module_name in support_dict, Exception(
        "metric only support {}".format(support_dict))
    module_class = eval(module_name)(**config)
    return module_class



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽

import re


class JSCommentRemover:
    """ js代码注释删除器，使用了比较严谨的规则，但也不能说特别完美 """

    def __init__(self, js_source_code):
        self.js_source_code = js_source_code + '\n\n             '  # Adding buffer to prevent out-of-bound errors
        self.js_code_length = len(js_source_code)
        self.index_pointer = -1
        self.comment_placeholder = ''  # Placeholder for comments
        self.js_without_comment = ''

    def _handle_normal_quote(self, delimiter='"'):
        while self.index_pointer < self.js_code_length:
            self.index_pointer += 1
            char = self.js_source_code[self.index_pointer]
            self.js_without_comment += char

            if char == delimiter:
                return
            elif char == '\\':
                self.index_pointer += 1
                if self.js_source_code[self.index_pointer + 1] == delimiter:
                    pass
                self.js_without_comment += self.js_source_code[self.index_pointer]
            elif char == '\n':
                raise ValueError(f'Syntax Error: Missing right string delimiter: {delimiter}')

    def _handle_template_string(self, delimiter='`'):
        while self.index_pointer < self.js_code_length:
            self.index_pointer += 1
            char = self.js_source_code[self.index_pointer]
            self.js_without_comment += char

            if char == delimiter:
                return
            elif char == '\\':
                self.index_pointer += 1
                self.js_without_comment += self.js_source_code[self.index_pointer]
            elif char == '$' and self.js_source_code[self.index_pointer + 1] == '{':
                self.index_pointer += 1
                self.js_without_comment += '{'
                self._nested_expression_inside_template_string()

    def _nested_expression_inside_template_string(self):
        while self.index_pointer < self.js_code_length:
            self.index_pointer += 1
            char = self.js_source_code[self.index_pointer]
            if char == '}':
                self.js_without_comment += char
                return
            elif char in '"\'':
                self.js_without_comment += char
                self._handle_normal_quote(delimiter=char)
            elif char == "`":
                self.js_without_comment += char
                self._handle_template_string()
            elif char == '/':
                self._handle_slash()

    def _handle_slash(self):
        next_char = self.js_source_code[self.index_pointer + 1]
        if next_char == '/':
            self._single_line_comment()
            self.js_without_comment += self.comment_placeholder
        elif next_char == '*':
            self.js_without_comment += self.comment_placeholder
            self.index_pointer += 1
            self._multi_line_comment()
        elif bool(re.search(r'[\n\r(\[\{=+,;]([\t]*|\\[\n\r])*[\t]*$', self.js_without_comment)):
            self.js_without_comment += '/'
            self._regular_expression()
        else:
            self.js_without_comment += '/'

    def _single_line_comment(self):
        while self.js_source_code[self.index_pointer] not in '\n\r' and self.index_pointer < self.js_code_length:
            self.index_pointer += 1
        self.js_without_comment += '\n'

    def _multi_line_comment(self):
        while not (
                self.js_source_code[self.index_pointer] == '*' and self.js_source_code[self.index_pointer + 1] == '/'):
            self.index_pointer += 1
        self.index_pointer += 1

    def _regular_expression(self):
        while self.js_source_code[self.index_pointer + 1] != '/':
            self.index_pointer += 1
            self.js_without_comment += self.js_source_code[self.index_pointer]
            if self.js_source_code[self.index_pointer] == '\\':
                self.index_pointer += 1
                self.js_without_comment += self.js_source_code[self.index_pointer]

    def remove_comments(self):
        while self.index_pointer < self.js_code_length:
            self.index_pointer += 1
            char = self.js_source_code[self.index_pointer]
            if char == '/':
                self._handle_slash()
                continue
            elif char in '"\'':
                self.js_without_comment += char
                self._handle_normal_quote(delimiter=char)
            elif char == "`":
                self.js_without_comment += char
                self._handle_template_string()
            elif char == "S" and self.js_code_length - self.index_pointer > 11 and \
                    self.js_source_code[self.index_pointer:self.index_pointer + 11] == 'String.raw`':
                self.js_without_comment += char
                self.index_pointer += 10
                self.js_without_comment += 'tring.raw`'
                self._handle_template_string()
            else:
                self.js_without_comment += char
        return self.js_without_comment


def remove_js_comments(js_source_code):
    remover = JSCommentRemover(js_source_code)
    return remover.remove_comments()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2023/08/02 14:05

import os
import random
import re
import sys
import json
import time
from types import SimpleNamespace

from PyQt5.QtWidgets import QApplication, QMainWindow, QAction, QFileDialog, QListWidget, QLineEdit, QVBoxLayout, \
    QSplitter, QTreeView, QPlainTextEdit, QPushButton, QLabel, QHBoxLayout, QSizePolicy, QWidget, QStatusBar, \
    QAbstractItemView, QHeaderView, QMessageBox
from PyQt5.QtWidgets import QItemDelegate, QTextEdit
from PyQt5.QtWidgets import QItemDelegate, QDialog, QVBoxLayout, QTextEdit, QPushButton

from PyQt5.QtGui import QTextOption, QStandardItemModel, QStandardItem
from PyQt5.QtCore import Qt, QModelIndex, QSettings, QFileInfo

from pyxllib.file.specialist import XlPath

# 一个专门存储大字符串的命名空间
LargeStrings = SimpleNamespace()

# 命名并存储QTreeView的样式（旧的格式配置，在新版中已经不起作用）
LargeStrings.treeViewStyles = """
QTreeView::item {  /* 设置网格线 */
    border: 1px solid black;
}

QTreeView::item:selected {
    background: black;
    color: white;
}

QTreeView::item:selected:active {
    background: black;
    color: white;
}

QTreeView::item:selected:!active {
    background: black;
    color: white;
}
""".strip()


class MyTreeView(QTreeView):
    def __init__(self, parent=None):
        super().__init__(parent)

    def edit(self, index, trigger, event):
        if trigger == QAbstractItemView.DoubleClicked:
            return False
            # 如果是第 0 列 (keys)，则禁止编辑
        if index.column() == 0:
            return False
        return super().edit(index, trigger, event)


class KeyStandardItem(QStandardItem):
    def data(self, role=None):
        if role == Qt.TextAlignmentRole:
            return Qt.AlignLeft | Qt.AlignVCenter
        return super().data(role)


class ExpandedTextEditDelegate(QItemDelegate):
    def createEditor(self, parent, option, index):
        return QTextEdit(parent)

    def sizeHint(self, option, index):
        size = super().sizeHint(option, index)
        return size

    def setEditorData(self, editor, index):
        value = index.model().data(index, Qt.EditRole)
        editor.setPlainText(value)

    def setModelData(self, editor, model, index):
        value = editor.toPlainText()
        model.setData(index, value, Qt.EditRole)

    def updateEditorGeometry(self, editor, option, index):
        editor.setGeometry(option.rect)


class CompactTextEditDelegate(ExpandedTextEditDelegate):
    def sizeHint(self, option, index):
        size = super().sizeHint(option, index)
        # todo 如果一个4k的屏幕使用2k的时候，这个缩放比例不会自动兼容，阁下又当如何应对
        size.setHeight(20)  # 限制最大高度为20像素，限定每个条目只展示一行
        return size

    def paint(self, painter, option, index):
        text = index.model().data(index)

        # 只显示前100个字符
        text = text[:100] + '...' if len(text) > 100 else text

        painter.drawText(option.rect, Qt.AlignLeft, text)


class JLineViewer(QMainWindow):
    def __init__(self):
        super(JLineViewer, self).__init__()
        self.load_settings()

        # 初始化 allItemsLoaded 变量
        self.allItemsLoaded = False

        self.initUI()
        # 开启部件接受拖放的能力（在windows中测试该功能失败）
        self.setAcceptDrops(True)

    def load_settings(self):
        self.settings = QSettings('pyxllib', 'JLineViewer')
        self.lastOpenDir = self.settings.value('lastOpenDir', '')

    def save_settings(self):
        self.settings.setValue('lastOpenDir', self.lastOpenDir)

    def initUI(self):
        self.listWidget = QListWidget()
        self.treeView = MyTreeView()
        self.plainTextEdit = QPlainTextEdit()
        self.searchLineEdit = QLineEdit()
        self.searchButton = QPushButton("普通搜索")

        self.listWidget.itemClicked.connect(self.loadJson)
        self.searchButton.clicked.connect(self.searchItems)
        self.searchLineEdit.returnPressed.connect(self.searchItems)
        self.treeView.clicked.connect(self.editItem)

        splitter = QSplitter(Qt.Horizontal)
        splitter.addWidget(self.addPane(self.listWidget, 'JSONL Items'))
        splitter.addWidget(self.addPane(self.treeView, 'JSON Tree View'))
        splitter.addWidget(self.addPane(self.plainTextEdit, 'Selected Content'))
        splitter.setSizes([100, 300, 200])

        layout = QVBoxLayout()
        searchLayout = QHBoxLayout()
        searchLayout.addWidget(QLabel("搜索条目："))
        searchLayout.addWidget(self.searchLineEdit)
        searchLayout.addWidget(self.searchButton)
        layout.addLayout(searchLayout)
        layout.addWidget(splitter)

        self.regexSearchButton = QPushButton("正则搜索")
        self.regexSearchButton.clicked.connect(self.regexSearchItems)  # 连接新的槽函数
        searchLayout.addWidget(self.regexSearchButton)  # 添加按钮到布局

        centralWidget = QWidget()
        self.setCentralWidget(centralWidget)
        centralWidget.setLayout(layout)

        openFile = QAction('打开文件', self)
        openFile.setShortcut('Ctrl+O')
        openFile.setStatusTip('打开新文件，可以打开jsonl或json格式的文件')
        openFile.triggered.connect(self.showDialog)

        self.loadAllButton = QPushButton("加载全部")
        self.loadAllButton.setStatusTip('对jsonl最多只会预加载1000行，点击该按钮可以加载剩余全部条目')
        self.loadAllButton.clicked.connect(self.loadAllItems)

        saveFile = QAction('保存文件', self)
        saveFile.setShortcut('Ctrl+S')
        saveFile.setStatusTip('保存文件')
        saveFile.triggered.connect(self.saveFile)

        toolbar = self.addToolBar('文件')
        toolbar.addAction(openFile)
        toolbar.addWidget(self.loadAllButton)  # 将按钮添加到布局中
        # toolbar.addAction(saveFile)

        self.statusBar = QStatusBar()
        self.setStatusBar(self.statusBar)

        self.setGeometry(300, 300, 350, 300)
        self.setWindowTitle('JLineEditor')
        self.treeView.setAlternatingRowColors(True)
        self.treeView.setIndentation(20)
        # self.treeView.setSortingEnabled(True)
        self.treeView.setStyleSheet(LargeStrings.treeViewStyles)
        # self.plainTextEdit.setWordWrapMode(QTextOption.WordWrap)
        self.plainTextEdit.setReadOnly(True)
        self.showMaximized()

        self.treeView.setSortingEnabled(False)  # 禁止排序
        self.treeView.setAnimated(False)
        # self.plainTextEdit.textChanged.connect(self.updateJson)  # 连接 textChanged 信号到新的槽函数

        self.delegate_mode = 'compact'
        self.toggleDelegateButton = QAction('切换显示模式', self)
        self.toggleDelegateButton.triggered.connect(self.toggleDelegate)
        toolbar.addAction(self.toggleDelegateButton)

    def toggleDelegate(self):
        if self.delegate_mode == 'compact':
            self.treeView.setItemDelegate(ExpandedTextEditDelegate(self.treeView))
            self.toggleDelegateButton.setText('切换为紧凑模式')
            self.delegate_mode = 'expanded'
        else:
            self.treeView.setItemDelegate(CompactTextEditDelegate(self.treeView))
            self.toggleDelegateButton.setText('切换为扩展模式')
            self.delegate_mode = 'compact'

    def addPane(self, widget, title):
        layout = QVBoxLayout()
        layout.addWidget(QLabel(title))
        layout.addWidget(widget)
        pane = QWidget()
        pane.setLayout(layout)
        return pane

    def saveFile(self):
        fname = QFileDialog.getSaveFileName(self, '保存文件',
                                            self.lastOpenDir if hasattr(self, 'lastOpenPath') else '/home')

        if fname[0]:
            self.lastOpenDir = fname[0]
            with open(fname[0], 'w', encoding='utf8') as f:
                for line in self.lines:
                    f.write(line)

    def updateJson(self):
        newText = self.plainTextEdit.toPlainText()
        self.currentlyEditingItem.setText(newText)  # 更新模型项的内容

        # 更新 JSON 数据
        # self.lines[self.listWidget.currentRow()] = self.modelToJson(self.model)

    def showDialog(self, *, fname=None):
        if fname is None:
            fname = QFileDialog.getOpenFileName(self, '打开文件',
                                                self.lastOpenDir,
                                                "JSON files (*.json *.jsonl)")
            fname = fname[0]

        if fname:
            self.lastOpenDir = os.path.dirname(QFileInfo(fname).absolutePath())
            self.save_settings()

            # 打开新文件时，重置 allItemsLoaded 变量
            self.allItemsLoaded = False

            # 清空旧数据
            self.lines = []
            self.listWidget.clear()

            # 1 打开文件
            start_time = time.time()  # 开始计时
            self.lastOpenDir = fname
            if fname.endswith('.json'):
                with open(fname, 'r', encoding='utf8') as f:
                    jsonData = json.load(f)
                    self.lines = [json.dumps(jsonData, ensure_ascii=False)]
            else:
                with open(fname, 'r', encoding='utf8') as f:
                    self.lines = f.readlines()
            self.statusBar.showMessage(f"文件打开耗时: {time.time() - start_time:.2f} 秒")
            QApplication.processEvents()

            # 2 加载条目数据
            start_time = time.time()
            self.setWindowTitle(f'JLineViewer - {fname}')
            self.listWidget.addItems([f'{i + 1}. {line.strip()}' for i, line in enumerate(self.lines[:1000])])
            QApplication.processEvents()

            # TODO 只有总条目数大于1000时才显示"仅预加载1000条"
            if len(self.lines) > 1000:
                self.statusBar.showMessage(f"总条目数: {len(self.lines)}, 仅预加载1000条，"
                                           f"加载条目耗时: {time.time() - start_time:.2f} 秒")
            else:
                self.statusBar.showMessage(f"总条目数: {len(self.lines)}, 加载条目耗时: {time.time() - start_time:.2f} 秒")

    def loadJson(self, item):
        index = self.listWidget.row(item)
        jsonData = json.loads(self.lines[index])
        self.model = self.dictToModel(jsonData)
        self.treeView.setModel(self.model)
        self.treeView.expandAll()

        # 使用自定义的delegate
        if self.delegate_mode == 'compact':
            self.treeView.setItemDelegate(CompactTextEditDelegate(self.treeView))
        else:
            self.treeView.setItemDelegate(ExpandedTextEditDelegate(self.treeView))

        self.treeView.header().setSectionResizeMode(0, QHeaderView.ResizeToContents)
        self.treeView.header().setSectionResizeMode(1, QHeaderView.Stretch)

        if hasattr(self, 'lastClickedPath'):
            path = self.lastClickedPath
            index = self.model.index(path[0][0], path[0][1])  # 从根开始
            for row, col in path[1:]:
                index = self.model.index(row, col, index)
            if index.isValid():
                self.treeView.setCurrentIndex(index)
                self.editItem(index)

    def loadAllItems(self):
        if not self.allItemsLoaded:
            start_time = time.time()
            self.listWidget.addItems([f'{i + 1}. {line.strip()}' for i, line in enumerate(self.lines[1000:])])
            QApplication.processEvents()
            self.statusBar.showMessage(f"全部加载完毕, 总条目数: {len(self.lines)}, 加载耗时: {time.time() - start_time:.2f} 秒")

            # 加载完所有项目后，设置 allItemsLoaded 变量为 True
            self.allItemsLoaded = True
        else:
            self.statusBar.showMessage(f"所有条目已经加载完毕。")

    def searchItems(self):
        if hasattr(self, 'lines'):
            start_time = time.time()
            searchText = self.searchLineEdit.text()
            foundCount = 0
            for i in range(self.listWidget.count()):
                item = self.listWidget.item(i)
                if searchText in item.text():
                    item.setHidden(False)
                    foundCount += 1
                else:
                    item.setHidden(True)
            QApplication.processEvents()
            self.statusBar.showMessage(
                f"总条目数: {len(self.lines)}, 找到: {foundCount}, 搜索耗时: {time.time() - start_time:.2f} 秒")

    def regexSearchItems(self):
        if hasattr(self, 'lines'):
            start_time = time.time()
            searchText = self.searchLineEdit.text()
            regexPattern = re.compile(searchText)  # 使用输入的文本创建正则表达式
            foundCount = 0
            for i in range(self.listWidget.count()):
                item = self.listWidget.item(i)
                if regexPattern.search(item.text()):  # 使用正则表达式搜索
                    item.setHidden(False)
                    foundCount += 1
                else:
                    item.setHidden(True)
            self.statusBar.showMessage(
                f"总条目数: {len(self.lines)}, 找到: {foundCount}, 搜索耗时: {time.time() - start_time:.2f} 秒")

    def itemToData(self, key_item):
        value_item = key_item.model().item(key_item.row(), 1)
        return value_item.text()

    def editItem(self, index=None):
        self.currentlyEditingItem = self.model.itemFromIndex(index)  # 保存当前正在编辑的项
        target_text = self.currentlyEditingItem.text()
        self.plainTextEdit.setPlainText(target_text)

        # 保存路径
        self.lastClickedPath = []
        while index.isValid():
            self.lastClickedPath.append((index.row(), index.column()))
            index = index.parent()
        self.lastClickedPath.reverse()

    def dictToModel(self, data, parent=None):
        if parent is None:
            parent = QStandardItemModel()
            parent.setHorizontalHeaderLabels(['Key', 'Value'])
        if isinstance(data, dict):
            for key, value in data.items():
                self.dataToModel(key, value, parent)
        return parent

    def dataToModel(self, key, value, parent):
        if isinstance(value, dict):
            item = KeyStandardItem(key)
            parent.appendRow([item, QStandardItem('')])
            for k, v in value.items():
                self.dataToModel(k, v, item)
        elif isinstance(value, list):
            # 方案 1
            # for i, v in enumerate(value):
            #     self.dataToModel(f"{key}[{i}]", v, parent)

            # 方案2 添加一个父节点
            list_parent = KeyStandardItem(key)
            parent.appendRow([list_parent, QStandardItem('')])
            # 将list的元素添加到父节点下
            for i, v in enumerate(value):
                self.dataToModel(f"{key}[{i}]", v, list_parent)
        else:
            parent.appendRow([KeyStandardItem(key), QStandardItem(str(value))])

    # def itemChanged(self, item):
    #     # 当一个模型项改变时，重新生成 JSON 数据
    #     self.lines[self.listWidget.currentRow()] = self.modelToJson(self.model)
    #
    # def modelToJson(self, model, parent=QModelIndex(), key=None):
    #     """ 这段功能有问题，暂不能开启 """
    #     rows = model.rowCount(parent)
    #     if rows == 0:
    #         # leaf node
    #         sibling = model.sibling(parent.row(), 1, parent)
    #         return model.data(sibling)
    #     else:
    #         # branch node
    #         json_data = {}
    #         for i in range(rows):
    #             index = model.index(i, 0, parent)
    #             child_key = model.data(index)
    #             child_value = self.modelToJson(model, index, child_key)
    #             json_data[child_key] = child_value
    #         return json.dumps(json_data, ensure_ascii=False)

    def dragEnterEvent(self, event):
        """
        当用户开始拖动文件到部件上时，这个方法会被调用。
        我们需要检查拖动的数据是不是文件类型（mime类型是'text/uri-list'）。
        """
        print("dragEnterEvent called")
        # if event.mimeData().hasFormat('text/uri-list'):
        event.acceptProposedAction()

    def dropEvent(self, event):
        """
        当用户在部件上释放（drop）文件时，这个方法会被调用。
        我们需要获取文件路径，然后判断文件类型是不是我们支持的类型。
        如果文件是我们支持的类型，我们就可以处理这个文件。
        """
        print("dropEvent called")
        # 获取文件路径
        file_paths = event.mimeData().urls()
        if file_paths:
            file_path = file_paths[0].toLocalFile()  # 取第一个文件
            # 检查文件扩展名是不是我们支持的类型
            if file_path.endswith(('.json', '.jsonl')):
                # 调用处理文件的方法
                self.showDialog(fname=file_path)
            else:
                QMessageBox.warning(self, "File Type Error",
                                    "Only .json or .jsonl files are supported.")

    # 当应用程序关闭时，保存设置
    def closeEvent(self, event):
        self.save_settings()
        event.accept()


def start_jlineviewer(fname=None):
    app = QApplication(sys.argv)
    ex = JLineViewer()
    if isinstance(fname, list):  # 可以输入一个list字典数据，会转存到临时目录里查看
        tempfile = XlPath.tempfile(suffix='.jsonl')
        tempfile.write_jsonl(fname)
        fname = tempfile
    if fname:
        ex.showDialog(fname=fname)
    sys.exit(app.exec_())


if __name__ == '__main__':
    app = QApplication(sys.argv)
    ex = JLineViewer()

    if len(sys.argv) > 1:
        ex.showDialog(fname=sys.argv[1])

    sys.exit(app.exec_())



# copyright (c) 2019 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import paddle
from paddle import nn
import paddle.nn.functional as F
from paddle import ParamAttr


class ConvBNLayer(nn.Layer):
    def __init__(self,
                 in_channels,
                 out_channels,
                 kernel_size,
                 stride,
                 groups=1,
                 if_act=True,
                 act=None,
                 name=None):
        super(ConvBNLayer, self).__init__()
        self.if_act = if_act
        self.act = act
        self.conv = nn.Conv2D(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=(kernel_size - 1) // 2,
            groups=groups,
            weight_attr=ParamAttr(name=name + '_weights'),
            bias_attr=False)
  
        self.bn = nn.BatchNorm(
            num_channels=out_channels,
            act=act,
            param_attr=ParamAttr(name="bn_" + name + "_scale"),
            bias_attr=ParamAttr(name="bn_" + name + "_offset"),
            moving_mean_name="bn_" + name + "_mean",
            moving_variance_name="bn_" + name + "_variance")

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        return x


class DeConvBNLayer(nn.Layer):
    def __init__(self,
                 in_channels,
                 out_channels,
                 kernel_size,
                 stride,
                 groups=1,
                 if_act=True,
                 act=None,
                 name=None):
        super(DeConvBNLayer, self).__init__()
        self.if_act = if_act
        self.act = act
        self.deconv = nn.Conv2DTranspose(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=(kernel_size - 1) // 2,
            groups=groups,
            weight_attr=ParamAttr(name=name + '_weights'),
            bias_attr=False)
        self.bn = nn.BatchNorm(
            num_channels=out_channels,
            act=act,
            param_attr=ParamAttr(name="bn_" + name + "_scale"),
            bias_attr=ParamAttr(name="bn_" + name + "_offset"),
            moving_mean_name="bn_" + name + "_mean",
            moving_variance_name="bn_" + name + "_variance")

    def forward(self, x):
        x = self.deconv(x)
        x = self.bn(x)
        return x


class FPN_Up_Fusion(nn.Layer):
    def __init__(self, in_channels):
        super(FPN_Up_Fusion, self).__init__()
        in_channels = in_channels[::-1]
        out_channels = [256, 256, 192, 192, 128]
                
        self.h0_conv = ConvBNLayer(in_channels[0], out_channels[0], 1, 1, act=None, name='fpn_up_h0')
        self.h1_conv = ConvBNLayer(in_channels[1], out_channels[1], 1, 1, act=None, name='fpn_up_h1')
        self.h2_conv = ConvBNLayer(in_channels[2], out_channels[2], 1, 1, act=None, name='fpn_up_h2')
        self.h3_conv = ConvBNLayer(in_channels[3], out_channels[3], 1, 1, act=None, name='fpn_up_h3')
        self.h4_conv = ConvBNLayer(in_channels[4], out_channels[4], 1, 1, act=None, name='fpn_up_h4')

        self.g0_conv = DeConvBNLayer(out_channels[0], out_channels[1], 4, 2, act=None, name='fpn_up_g0')

        self.g1_conv = nn.Sequential(
            ConvBNLayer(out_channels[1], out_channels[1], 3, 1, act='relu', name='fpn_up_g1_1'),
            DeConvBNLayer(out_channels[1], out_channels[2], 4, 2, act=None, name='fpn_up_g1_2')
        )
        self.g2_conv = nn.Sequential(
            ConvBNLayer(out_channels[2], out_channels[2], 3, 1, act='relu', name='fpn_up_g2_1'),
            DeConvBNLayer(out_channels[2], out_channels[3], 4, 2, act=None, name='fpn_up_g2_2')
        )
        self.g3_conv = nn.Sequential(
            ConvBNLayer(out_channels[3], out_channels[3], 3, 1, act='relu', name='fpn_up_g3_1'),
            DeConvBNLayer(out_channels[3], out_channels[4], 4, 2, act=None, name='fpn_up_g3_2')
        )

        self.g4_conv = nn.Sequential(
            ConvBNLayer(out_channels[4], out_channels[4], 3, 1, act='relu', name='fpn_up_fusion_1'),
            ConvBNLayer(out_channels[4], out_channels[4], 1, 1, act=None, name='fpn_up_fusion_2')
        )

    def _add_relu(self, x1, x2):
        x = paddle.add(x=x1, y=x2)
        x = F.relu(x)
        return x

    def forward(self, x):
        f = x[2:][::-1]
        h0 = self.h0_conv(f[0])
        h1 = self.h1_conv(f[1])
        h2 = self.h2_conv(f[2])
        h3 = self.h3_conv(f[3])
        h4 = self.h4_conv(f[4])

        g0 = self.g0_conv(h0)
        g1 = self._add_relu(g0, h1)
        g1 = self.g1_conv(g1)
        g2 = self.g2_conv(self._add_relu(g1, h2))
        g3 = self.g3_conv(self._add_relu(g2, h3))
        g4 = self.g4_conv(self._add_relu(g3, h4))

        return g4


class FPN_Down_Fusion(nn.Layer):
    def __init__(self, in_channels):
        super(FPN_Down_Fusion, self).__init__()
        out_channels = [32, 64, 128]

        self.h0_conv = ConvBNLayer(in_channels[0], out_channels[0], 3, 1, act=None, name='fpn_down_h0')
        self.h1_conv = ConvBNLayer(in_channels[1], out_channels[1], 3, 1, act=None, name='fpn_down_h1')
        self.h2_conv = ConvBNLayer(in_channels[2], out_channels[2], 3, 1, act=None, name='fpn_down_h2')

        self.g0_conv = ConvBNLayer(out_channels[0], out_channels[1], 3, 2, act=None, name='fpn_down_g0')

        self.g1_conv = nn.Sequential(
            ConvBNLayer(out_channels[1], out_channels[1], 3, 1, act='relu', name='fpn_down_g1_1'),
            ConvBNLayer(out_channels[1], out_channels[2], 3, 2, act=None, name='fpn_down_g1_2')            
        )

        self.g2_conv = nn.Sequential(
            ConvBNLayer(out_channels[2], out_channels[2], 3, 1, act='relu', name='fpn_down_fusion_1'),
            ConvBNLayer(out_channels[2], out_channels[2], 1, 1, act=None, name='fpn_down_fusion_2')            
        )

    def forward(self, x):
        f = x[:3]
        h0 = self.h0_conv(f[0])
        h1 = self.h1_conv(f[1])
        h2 = self.h2_conv(f[2])
        g0 = self.g0_conv(h0)
        g1 = paddle.add(x=g0, y=h1)
        g1 = F.relu(g1)
        g1 = self.g1_conv(g1)
        g2 = paddle.add(x=g1, y=h2)
        g2 = F.relu(g2)
        g2 = self.g2_conv(g2)
        return g2


class Cross_Attention(nn.Layer):
    def __init__(self, in_channels):
        super(Cross_Attention, self).__init__()
        self.theta_conv = ConvBNLayer(in_channels, in_channels, 1, 1, act='relu', name='f_theta')
        self.phi_conv = ConvBNLayer(in_channels, in_channels, 1, 1, act='relu', name='f_phi')
        self.g_conv = ConvBNLayer(in_channels, in_channels, 1, 1, act='relu', name='f_g')

        self.fh_weight_conv = ConvBNLayer(in_channels, in_channels, 1, 1, act=None, name='fh_weight')
        self.fh_sc_conv = ConvBNLayer(in_channels, in_channels, 1, 1, act=None, name='fh_sc')

        self.fv_weight_conv = ConvBNLayer(in_channels, in_channels, 1, 1, act=None, name='fv_weight')
        self.fv_sc_conv = ConvBNLayer(in_channels, in_channels, 1, 1, act=None, name='fv_sc')

        self.f_attn_conv = ConvBNLayer(in_channels * 2, in_channels, 1, 1, act='relu', name='f_attn')

    def _cal_fweight(self, f, shape):
        f_theta, f_phi, f_g = f
        #flatten
        f_theta = paddle.transpose(f_theta, [0, 2, 3, 1])
        f_theta = paddle.reshape(f_theta, [shape[0] * shape[1], shape[2], 128])
        f_phi = paddle.transpose(f_phi, [0, 2, 3, 1])
        f_phi = paddle.reshape(f_phi, [shape[0] * shape[1], shape[2], 128])
        f_g = paddle.transpose(f_g, [0, 2, 3, 1])
        f_g = paddle.reshape(f_g, [shape[0] * shape[1], shape[2], 128])
        #correlation
        f_attn = paddle.matmul(f_theta, paddle.transpose(f_phi, [0, 2, 1]))
        #scale
        f_attn = f_attn / (128**0.5)
        f_attn = F.softmax(f_attn)
        #weighted sum
        f_weight = paddle.matmul(f_attn, f_g)
        f_weight = paddle.reshape(
            f_weight, [shape[0], shape[1], shape[2], 128])
        return f_weight

    def forward(self, f_common):
        f_shape = paddle.shape(f_common)
        # print('f_shape: ', f_shape)

        f_theta = self.theta_conv(f_common)
        f_phi = self.phi_conv(f_common)
        f_g = self.g_conv(f_common)

        ######## horizon ########
        fh_weight = self._cal_fweight([f_theta, f_phi, f_g], 
                                        [f_shape[0], f_shape[2], f_shape[3]])
        fh_weight = paddle.transpose(fh_weight, [0, 3, 1, 2])
        fh_weight = self.fh_weight_conv(fh_weight)
        #short cut
        fh_sc = self.fh_sc_conv(f_common)
        f_h = F.relu(fh_weight + fh_sc)

        ######## vertical ########
        fv_theta = paddle.transpose(f_theta, [0, 1, 3, 2])
        fv_phi = paddle.transpose(f_phi, [0, 1, 3, 2])
        fv_g = paddle.transpose(f_g, [0, 1, 3, 2])
        fv_weight = self._cal_fweight([fv_theta, fv_phi, fv_g], 
                                        [f_shape[0], f_shape[3], f_shape[2]])
        fv_weight = paddle.transpose(fv_weight, [0, 3, 2, 1])
        fv_weight = self.fv_weight_conv(fv_weight)
        #short cut
        fv_sc = self.fv_sc_conv(f_common)
        f_v = F.relu(fv_weight + fv_sc)

        ######## merge ########
        f_attn = paddle.concat([f_h, f_v], axis=1)
        f_attn = self.f_attn_conv(f_attn)
        return f_attn


class SASTFPN(nn.Layer):
    def __init__(self, in_channels, with_cab=False, **kwargs):
        super(SASTFPN, self).__init__()
        self.in_channels = in_channels
        self.with_cab = with_cab
        self.FPN_Down_Fusion = FPN_Down_Fusion(self.in_channels)
        self.FPN_Up_Fusion = FPN_Up_Fusion(self.in_channels)
        self.out_channels = 128
        self.cross_attention = Cross_Attention(self.out_channels)

    def forward(self, x):
        #down fpn
        f_down = self.FPN_Down_Fusion(x)

        #up fpn
        f_up = self.FPN_Up_Fusion(x)

        #fusion
        f_common = paddle.add(x=f_down, y=f_up)
        f_common = F.relu(f_common)

        if self.with_cab:
            # print('enhence f_common with CAB.')
            f_common = self.cross_attention(f_common)

        return f_common
